import logging
from typing import TYPE_CHECKING, List

from botocore.exceptions import ClientError

from aibs_informatics_aws_utils.core import AWSService
from aibs_informatics_aws_utils.exceptions import AWSError

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_ecs.type_defs import (
        DescribeContainerInstancesRequestTypeDef,
        DescribeContainerInstancesResponseTypeDef,
    )
else:
    DescribeContainerInstancesRequestTypeDef = dict
    DescribeContainerInstancesResponseTypeDef = dict


logger = logging.getLogger(__name__)

get_ecs_client = AWSService.ECS.get_client


def ecs_describe_container_instances(
    container_instances: List[str], cluster: str, **kwargs
) -> DescribeContainerInstancesResponseTypeDef:
    ecs = get_ecs_client()
    try:
        return ecs.describe_container_instances(
            containerInstances=container_instances, cluster=cluster, **kwargs
        )
    except ClientError as e:
        msg = (
            "Error retrieving container instance metadata for container "
            f"instances={container_instances}, cluster={cluster}: {e}"
        )
        logger.error(msg, exc_info=True)
        raise AWSError(msg)



import hashlib
import logging
from datetime import datetime
from typing import TYPE_CHECKING, List, Optional, Union

from aibs_informatics_core.env import EnvBase
from aibs_informatics_core.models.aws.sfn import ExecutionArn as _ExecutionArn
from aibs_informatics_core.models.aws.sfn import StateMachineArn as _StateMachineArn
from aibs_informatics_core.utils.time import get_current_time
from botocore.exceptions import ClientError

from aibs_informatics_aws_utils.core import (
    AWSService,
    get_account_id,
    get_client_error_code,
    get_client_error_message,
    get_region,
)
from aibs_informatics_aws_utils.exceptions import (
    AWSError,
    InvalidAmazonResourceNameError,
    ResourceNotFoundError,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_stepfunctions.literals import ExecutionStatusType, IncludedDataType
    from mypy_boto3_stepfunctions.type_defs import (
        DescribeExecutionOutputTypeDef,
        ExecutionListItemTypeDef,
        HistoryEventTypeDef,
        StateMachineListItemTypeDef,
        TaskFailedEventDetailsTypeDef,
        TaskSucceededEventDetailsTypeDef,
    )
else:
    IncludedDataType = str
    ExecutionStatusType = str
    DescribeExecutionOutputTypeDef = dict
    HistoryEventTypeDef = dict
    DescribeExecutionOutputTypeDef = dict
    ExecutionStatusType = str
    ExecutionListItemTypeDef = dict
    HistoryEventTypeDef = dict
    StateMachineListItemTypeDef = dict
    TaskFailedEventDetailsTypeDef = dict
    TaskSucceededEventDetailsTypeDef = dict


get_sfn_client = AWSService.STEPFUNCTIONS.get_client


class StateMachineArn(_StateMachineArn):
    @classmethod
    def from_components(  # type: ignore
        cls,
        state_machine_name: str,
        region: Optional[str] = None,
        account_id: Optional[str] = None,
    ) -> "StateMachineArn":
        return StateMachineArn(
            ":".join(
                [
                    "arn",
                    "aws",
                    "states",
                    get_region(region),
                    account_id or get_account_id(),
                    "stateMachine",
                    state_machine_name,
                ]
            )
        )


class ExecutionArn(_ExecutionArn):
    @classmethod
    def from_components(  # type: ignore[override]
        cls,
        state_machine_name: str,
        execution_name: str,
        region: Optional[str] = None,
        account_id: Optional[str] = None,
    ) -> "ExecutionArn":
        return ExecutionArn(
            ":".join(
                [
                    "arn",
                    "aws",
                    "states",
                    get_region(region),
                    account_id or get_account_id(),
                    "execution",
                    state_machine_name,
                    execution_name,
                ]
            )
        )


# Note: Stepfunctions API start_execution is idempotent operation,
# so with same name and payload always returns the same execution, instead
# of duplicating for demand. This also means the same payload can only be
# executed once in an environment and the name cant be reused for 90 days.
def build_execution_name(payload: str, date: Optional[datetime] = None) -> str:
    try:
        str_to_encode = payload + date.isoformat() if date else ""
        return hashlib.sha256(str_to_encode.encode()).hexdigest()
    except TypeError as e:
        raise ValueError(f"JSON serialization failed for {payload}: {e}")
    except UnicodeDecodeError as e:
        raise ValueError(f"String Encoding failed for {payload}: {e}")
    except Exception as e:
        raise ValueError(f"Exception {e} raised for {payload}")


def get_execution_arn(
    state_machine_name: str,
    execution_name: str,
    env_base: Optional[EnvBase] = None,
    region: Optional[str] = None,
) -> ExecutionArn:
    sfn = get_sfn_client(region)
    state_machine = get_state_machine(name=state_machine_name, env_base=env_base, region=region)
    paginator = sfn.get_paginator("list_executions")
    iterator = paginator.paginate(
        stateMachineArn=state_machine["stateMachineArn"],
    )
    for list_executions_response in iterator:
        for execution in list_executions_response["executions"]:
            if execution["name"] == execution_name:
                return ExecutionArn(execution["executionArn"])
    else:
        raise InvalidAmazonResourceNameError(
            f"Could not find an execution ARN with "
            f"execution_name={execution_name}, "
            f"state_machine_name={state_machine['name']}"
        )


def describe_execution(
    execution_arn: str, included_data: IncludedDataType = "ALL_DATA", region: Optional[str] = None
) -> DescribeExecutionOutputTypeDef:
    sfn = get_sfn_client(region=get_region(region=region))

    execution_description = sfn.describe_execution(
        executionArn=execution_arn, includedData=included_data
    )
    return execution_description


def get_execution_history(
    execution_arn: Union[ExecutionArn, str],
    reverse_order: bool = False,
    include_execution_data: bool = False,
    region: Optional[str] = None,
) -> List[HistoryEventTypeDef]:
    sfn = get_sfn_client(region=get_region(region=region))
    execution_arn = ExecutionArn(execution_arn)
    paginator = sfn.get_paginator("get_execution_history")
    return [
        event
        for response in paginator.paginate(
            executionArn=execution_arn,
            reverseOrder=reverse_order,
            includeExecutionData=include_execution_data,
        )
        for event in response.get("events", [])
    ]


def start_execution(
    state_machine_name: str,
    state_machine_input: str,
    reuse_existing_execution: bool = False,
    execution_name: Optional[str] = None,
    env_base: Optional[EnvBase] = None,
    region: Optional[str] = None,
) -> ExecutionArn:
    """Starts a StepFn Execution

    Notes:
        * The state machine Arn is resolved using the name.
        * If ExecutionAlreadyExists and `reuse_existing_execution`=False,
          a unique execution name will be generated using
          aibs_informatics_core.utils.time.get_current_time().

    Args:
        state_machine_name (str): Name of state machine to execute
        state_machine_input (str): Serialized payload input for execution.
        reuse_existing_execution (bool, optional): Skip if existing execution submitted.
            Defaults to False.
        env_base (EnvBase, optional): environment base. Defaults to None.
        region (str, optional): ergion. Defaults to None.

    Returns:
        ExecutionArn: the execution arn
    """
    region = get_region(region=region)
    sfn = get_sfn_client()

    state_machine = get_state_machine(name=state_machine_name, env_base=env_base, region=region)
    state_machine_full_name = state_machine["name"]
    state_machine_arn = state_machine["stateMachineArn"]

    def _start_execution(execution_name: str):
        try:
            logger.info(
                f"Starting execution of step function {state_machine_arn} [{execution_name}]",
            )

            response = sfn.start_execution(
                stateMachineArn=state_machine_arn,
                name=execution_name,
                input=state_machine_input,
            )

        except ClientError as e:
            if get_client_error_code(e) == "ExecutionAlreadyExists":
                if reuse_existing_execution:
                    logger.info(f"ExecutionAlreadyExists for {state_machine_input} so skipping.")
                    return ExecutionArn.findall(get_client_error_message(e))[0]
                else:
                    return _start_execution(
                        build_execution_name(state_machine_input, date=get_current_time())
                    )
            elif get_client_error_code(e) == "StateMachineDoesNotExist":
                msg = (
                    f"State machine {state_machine_full_name} [{state_machine_arn}] does not exist"
                )
                logger.error(msg)
                raise ResourceNotFoundError(msg) from e
            else:
                raise AWSError(
                    f"Start StepFn failed with unknown ClientError for {state_machine_arn}: {e}"
                ) from e
        except Exception as e:
            raise AWSError(f"Unknown failure for {state_machine_arn}") from e
        return ExecutionArn(response["executionArn"])

    execution_name = execution_name or build_execution_name(state_machine_input)
    return _start_execution(execution_name)


def stop_execution(
    execution_arn: Union[ExecutionArn, str],
    error: Optional[str] = None,
    cause: Optional[str] = None,
    region: Optional[str] = None,
) -> datetime:
    sfn = get_sfn_client(region=get_region(region=region))
    response = sfn.stop_execution(
        executionArn=ExecutionArn(execution_arn), error=error or "", cause=cause or ""
    )
    return response["stopDate"]


def get_state_machine(
    name: str, env_base: Optional[EnvBase] = None, region: Optional[str] = None
) -> StateMachineListItemTypeDef:
    region = get_region(region=region)
    # env_base = env_base or get_env_base()

    matching_state_machines = [
        state_machine
        for state_machine in get_state_machines(env_base=env_base, region=region)
        if state_machine["name"].endswith(name)
    ]
    if len(matching_state_machines) == 0:
        raise ResourceNotFoundError(
            f"No state machines with env_base={env_base} and name suffix={name} exist"
        )
    elif len(matching_state_machines) > 1:
        raise AttributeError(
            f"More than 1 state machines with env_base={env_base} "
            f"and name suffix={name} exist: {matching_state_machines}"
        )
    return matching_state_machines[0]


def get_state_machines(
    env_base: Optional[EnvBase] = None, region: Optional[str] = None
) -> List[StateMachineListItemTypeDef]:
    region = get_region(region=region)
    sfn = get_sfn_client(region=region)
    paginator = sfn.get_paginator("list_state_machines")

    return [
        state_machine_info
        for response in paginator.paginate()
        for state_machine_info in response["stateMachines"]
        if not env_base or state_machine_info["name"].startswith(env_base)
    ]



import re
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Mapping, Match, Optional, Union, cast

from aibs_informatics_core.collections import ValidatedStr
from aibs_informatics_core.models.aws.dynamodb import (
    AttributeBaseExpression,
    ConditionBaseExpression,
    ConditionBaseExpressionString,
)
from aibs_informatics_core.utils.decorators import cached_property
from aibs_informatics_core.utils.logging import get_logger
from aibs_informatics_core.utils.modules import get_all_subclasses
from boto3.dynamodb.conditions import (
    Attr,
    AttributeBase,
    ConditionBase,
    ConditionExpressionBuilder,
    Key,
)
from boto3.dynamodb.types import TypeSerializer

logger = get_logger(__name__)


def condition_to_str(condition: Optional[ConditionBase]) -> Optional[str]:
    """Converts a ConditionBase Boto3 object to its str representation

    NOTE: Function should be removed if this PR is merged: https://github.com/boto/boto3/pull/3254

    Examples:

    >>> condition_to_str(Key("name").eq("new_name") & Attr("description").begins_with("new"))
    (name = new_name AND begins_with(description, new))

    >>> condition_to_str(Attr("description").contains("cool"))
    contains(description, cool)

    >>> condition_to_str(None)
    None
    """
    if condition is None:
        return None

    builder = ConditionExpressionBuilder()
    expression = builder.build_expression(condition)

    condition_expression = expression.condition_expression

    for name_placeholder, actual_name in expression.attribute_name_placeholders.items():
        condition_expression = condition_expression.replace(name_placeholder, str(actual_name))

    for value_placeholder, actual_value in expression.attribute_value_placeholders.items():
        condition_expression = condition_expression.replace(value_placeholder, str(actual_value))

    return condition_expression


@dataclass
class ExpressionComponentsBase:
    expression: str
    expression_attribute_names: Dict[str, str]
    expression_attribute_values: Dict[str, Any]

    @cached_property
    def expression_attribute_values__serialized(self) -> Dict[str, Dict[str, Any]]:
        serializer = TypeSerializer()
        return {
            k: cast(Dict[str, Any], serializer.serialize(v))
            for k, v in self.expression_attribute_values.items()
        }


@dataclass
class UpdateExpressionComponents(ExpressionComponentsBase):
    @property
    def update_expression(self) -> str:
        return self.expression

    @classmethod
    def from_dict(cls, attributes: Mapping[str, Any]) -> "UpdateExpressionComponents":
        expression_attr_names = {}
        expression_attr_values = {}
        update_expressions = []
        for i, (attr_name, attr_value) in enumerate(attributes.items()):
            attr_name_key = f"#n{i}"
            attr_value_key = f":v{i}"
            expression_attr_names[attr_name_key] = attr_name
            expression_attr_values[attr_value_key] = attr_value
            update_expressions.append(f"{attr_name_key} = {attr_value_key}")
        update_expression = "SET " + ", ".join(update_expressions)
        return UpdateExpressionComponents(
            expression=update_expression,
            expression_attribute_names=expression_attr_names,
            expression_attribute_values=expression_attr_values,
        )


@dataclass
class ConditionExpressionComponents(ExpressionComponentsBase):
    @property
    def condition_expression(self) -> str:
        return self.expression

    @classmethod
    def from_condition(
        cls, condition: ConditionBase, is_key_condition: bool
    ) -> "ConditionExpressionComponents":
        """Create ConditionExpressionComponents from a ConditionBase

        Args:
            condition (ConditionBase): The ConditionBase to create expression components from
            is_key_condition (bool): If the provided condition is for a query()
                KeyConditionExpression then this should be True. Otherwise if the condition is for
                a query() FilterExpression or for a put_item() ConditionExpression then this
                should be set to false.

        Returns:
            ConditionExpressionComponents
        """
        builder = ConditionExpressionBuilder()
        bce = builder.build_expression(condition, is_key_condition=is_key_condition)

        return ConditionExpressionComponents(
            expression=bce.condition_expression,
            expression_attribute_names=cast(Dict[str, str], bce.attribute_name_placeholders),
            expression_attribute_values=cast(Dict[str, Any], bce.attribute_value_placeholders),
        )

    def fix_collisions(
        self, other: "ConditionExpressionComponents"
    ) -> "ConditionExpressionComponents":
        """Modify other expression components such that no attribute name/values collide

        For example, given the following Expression components
            this:   ConditionExpressionComponents("#n0 = :v0", {"#n0": "key1"}, {":v0": {"S": "str_value"}})
            other:  ConditionExpressionComponents("#n0 = :v1", {"#n0": "key2"}, {":v1": {"S": "str_value"}})

        There is a collision of the attribute name '#n0' which would get converted to '#n1'
            output: ConditionExpressionComponents("#n1 = :v1", {"#n1": "key2"}, {":v1": {"S": "str_value"}})

        Args:
            other (ConditionExpressionComponents): the other Expression Condition to modify

        Returns:
            ConditionExpressionComponents: The other, as a modified non-overlapping Expression Condition
        """  # noqa: E501

        this_placeholder_names = AttrPlaceholder.sorted(self.expression_attribute_names.keys())
        this_placeholder_values = AttrPlaceholder.sorted(self.expression_attribute_values.keys())

        other_placeholder_names = AttrPlaceholder.sorted(other.expression_attribute_names.keys())
        other_placeholder_values = AttrPlaceholder.sorted(other.expression_attribute_values.keys())

        prefix_max_num_map = AttrPlaceholder.build_prefix_max_number_map(
            *this_placeholder_names,
            *this_placeholder_values,
            *other_placeholder_names,
            *other_placeholder_values,
        )

        # Colliding placeholders
        intersect_names = set(this_placeholder_names).intersection(other_placeholder_names)
        intersect_values = set(this_placeholder_values).intersection(other_placeholder_values)

        # create mapping for placeholder updates
        update_map: Dict[AttrPlaceholder, AttrPlaceholder] = {}

        # Now create new placeholders for colliding placeholders,
        # and add to the placeholder update mapping.
        for placeholder in intersect_names.union(intersect_values):
            prefix_max_num_map[placeholder.prefix] += 1
            update_map[placeholder] = AttrPlaceholder.from_components(
                placeholder.prefix, prefix_max_num_map[placeholder.prefix]
            )

        def _fetch(m: Match) -> AttrPlaceholder:
            ap = AttrPlaceholder(m.group(0))
            return update_map.get(ap, ap)

        # Now build the new expression components object
        return ConditionExpressionComponents(
            expression=re.sub(AttrPlaceholder.regex_pattern, _fetch, other.condition_expression),
            expression_attribute_names={
                update_map.get(AttrPlaceholder(p), p): other.expression_attribute_names[p]
                for p in other.expression_attribute_names
            },
            expression_attribute_values={
                update_map.get(AttrPlaceholder(p), p): other.expression_attribute_values[p]
                for p in other.expression_attribute_values
            },
        )


class AttrPlaceholder(ValidatedStr):
    regex_pattern = re.compile(r"([#:][a-zA-Z])([\d]+)")

    @property
    def prefix(self) -> str:
        return self.get_match_groups()[0]

    @property
    def number(self) -> int:
        return int(self.get_match_groups()[1])

    @classmethod
    def from_components(cls, prefix: str, number: int) -> "AttrPlaceholder":
        return cls(f"{prefix}{number}")

    @classmethod
    def sorted(
        cls, placeholders: Iterable[Union[str, "AttrPlaceholder"]]
    ) -> List["AttrPlaceholder"]:
        return sorted(map(AttrPlaceholder, placeholders), key=lambda _: _.number)

    @classmethod
    def build_prefix_max_number_map(cls, *placeholders: "AttrPlaceholder") -> Dict[str, int]:
        map: Dict[str, int] = defaultdict(int)
        for placeholder in placeholders:
            if map[placeholder.prefix] < placeholder.number:
                map[placeholder.prefix] = placeholder.number
        return map


class ConditionBaseTranslator:
    _ATTRIBUTE_BASE_CLASS_LOOKUP = {_.__name__: _ for _ in get_all_subclasses(AttributeBase)}
    _CONDITION_BASE_CLASS_LOOKUP = {
        (_.expression_format, _.expression_operator): _ for _ in get_all_subclasses(ConditionBase)
    }

    @classmethod
    def deserialize_attribute(
        cls, attribute_expression: AttributeBaseExpression
    ) -> Union[Key, Attr, AttributeBase]:
        return cls._ATTRIBUTE_BASE_CLASS_LOOKUP[attribute_expression.attr_class](
            attribute_expression.attr_name
        )

    @classmethod
    def deserialize_condition(
        cls, condition_expression: Union[str, ConditionBaseExpression], is_key: bool = True
    ) -> ConditionBase:
        """Convert ConditionBase Expression into ConditionBase

        Args:
            condition_expression (str|ConditionBaseExpression): The expression to convert.
            is_key (bool, optional): Used to infer attribute name if expression is string.
                Defaults to True.

        Returns:
            ConditionBase
        """

        def _deserialize_condition(ce: ConditionBaseExpression) -> ConditionBase:
            ce_key = (ce.format, ce.operator)
            condition_base_cls = cls._CONDITION_BASE_CLASS_LOOKUP[ce_key]
            ce_values: list[Union[AttributeBase, ConditionBase]] = []
            for ce_value in ce.values:
                if isinstance(ce_value, ConditionBaseExpression):
                    ce_values.append(_deserialize_condition(ce_value))
                elif isinstance(ce_value, AttributeBaseExpression):
                    ce_values.append(cls.deserialize_attribute(ce_value))
                else:
                    ce_values.append(ce_value)
            return condition_base_cls(*ce_values)

        if isinstance(condition_expression, str):
            condition_string = ConditionBaseExpressionString(condition_expression)
            condition_expression = condition_string.get_condition_expression(is_key=is_key)
        return _deserialize_condition(condition_expression)

    @classmethod
    def serialize_attribute(cls, attribute: AttributeBase) -> AttributeBaseExpression:
        return AttributeBaseExpression(attribute.__class__.__name__, attribute.name)

    @classmethod
    def serialize_condition(cls, condition: ConditionBase) -> ConditionBaseExpression:
        raw_expression = condition.get_expression()
        return ConditionBaseExpression(
            format=raw_expression["format"],
            operator=raw_expression["operator"],
            values=[
                cls.serialize_condition(value)
                if isinstance(value, ConditionBase)
                else (
                    cls.serialize_attribute(value) if isinstance(value, (AttributeBase)) else value
                )
                for value in raw_expression["values"]
            ],
        )

    @classmethod
    def _get_condition_base_operators(cls) -> List[str]:
        return [
            _.expression_operator
            for _ in ConditionBaseTranslator._CONDITION_BASE_CLASS_LOOKUP.values()
        ]



import hashlib
import json
import logging
import math
import os
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime, timezone
from enum import Enum
from functools import lru_cache
from multiprocessing.pool import ThreadPool
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Pattern,
    Set,
    Tuple,
    Union,
)
from urllib import parse

from aibs_informatics_core.models.aws.s3 import (
    S3URI,
    S3CopyRequest,
    S3DownloadRequest,
    S3PathStats,
    S3RestoreStatus,
    S3RestoreStatusEnum,
    S3StorageClass,
    S3TransferRequest,
    S3TransferResponse,
    S3UploadRequest,
)
from aibs_informatics_core.utils.decorators import retry
from aibs_informatics_core.utils.file_operations import (
    find_paths,
    get_path_with_root,
    remove_path,
    strip_path_root,
)
from aibs_informatics_core.utils.json import JSON
from aibs_informatics_core.utils.logging import get_logger
from aibs_informatics_core.utils.multiprocessing import parallel_starmap
from aibs_informatics_core.utils.time import get_current_time
from aibs_informatics_core.utils.tools.strtools import is_prefixed
from boto3.s3.transfer import TransferConfig as TransferConfig
from botocore.client import Config
from botocore.exceptions import (
    ClientError,
    ConnectionClosedError,
    EndpointConnectionError,
    ResponseStreamingError,
)

from aibs_informatics_aws_utils.core import (
    AWSService,
    client_error_code_check,
    get_client_error_message,
)
from aibs_informatics_aws_utils.exceptions import AWSError

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_s3 import S3Client
    from mypy_boto3_s3.service_resource import Object
else:
    S3Client = object
    Object = object


logger = get_logger()
logger.setLevel(level=logging.INFO)

get_s3_client = AWSService.S3.get_client
get_s3_resource = AWSService.S3.get_resource


S3_SCRATCH_TAGGING_KEY = "time-to-live"
S3_SCRATCH_TAGGING_VALUE = "scratch"


PathOrUri = Union[Path, S3URI]


SCRATCH_EXTRA_ARGS = {
    "Tagging": parse.urlencode({S3_SCRATCH_TAGGING_KEY: S3_SCRATCH_TAGGING_VALUE})
}

KB = 1024
MB = KB * KB
AWS_S3_DEFAULT_CHUNK_SIZE_BYTES = 8 * MB
LOCAL_ETAG_READ_BUFFER_BYTES = 8 * MB

# https://docs.aws.amazon.com/AmazonS3/latest/userguide/qfacts.html
AWS_S3_MULTIPART_LIMIT = 10000


def download_to_json_object(s3_path: S3URI, **kwargs) -> Dict[str, Any]:
    content = download_to_json(s3_path=s3_path, **kwargs)
    assert isinstance(content, dict)
    return content


def download_to_json(s3_path: S3URI, **kwargs) -> JSON:
    """Helper method to read a json file from S3."""
    logger.info(f"Reading file from s3: {s3_path}")
    s3_obj = get_object(s3_path=s3_path, **kwargs)

    try:
        data = json.load(s3_obj.get()["Body"])
    except Exception as e:
        raise AWSError(f"Error reading json data from {s3_path} [{e}]")

    return data


def download_s3_path(
    s3_path: S3URI,
    local_path: Path,
    exist_ok: bool = False,
    transfer_config: Optional[TransferConfig] = None,
    force: bool = True,
    size_only: bool = False,
    **kwargs,
):
    """Download an S3 Object or Folder to a local path

    This mimics the behavior of `aws s3 cp`, both with and without the `--recursive` flag.

    The logic for using recursive behavior is as follows:
    - for non-trailing-slash S3 paths,
        - If S3 path is an object, it will be downloaded as a file.
        - If S3 path is an object prefix, it will be downloaded as a folder.
    - for trailing-slash S3 paths,
        - If S3 path is a non-empty object and object prefix, an error will be raised.
        - If S3 path is an object prefix, it will be downloaded as a folder.
        - If S3 path is an object, it will be downloaded as a file.

    Args:
        s3_path (S3URI): URI of the object of folder
        local_path (Path): local destination path.
        exist_ok (bool, optional): If true, local path may exist previously. Defaults to False.

    Raises:
        ApplicationException: If S3 URI does not exist.
    """
    path_is_object = is_object(s3_path, **kwargs)
    path_is_prefix = is_object_prefix(s3_path, **kwargs)
    path_has_folder_slash = s3_path.key.endswith("/")
    path_is_non_empty_object_fn = lru_cache(None)(
        lambda: path_is_object and get_object(s3_path, **kwargs).content_length != 0
    )

    # Scenarios for object download:
    #   1. path IS OBJECT and NOT HAS FOLDER SLASH
    #   2. path IS OBJECT and HAS FOLDER SLASH but NOT OBJECT PREFIX
    # Scenarios for object prefix download:
    #   1. path IS OBJECT PREFIX and NOT OBJECT
    #   2. path IS OBJECT PREFIX and HAS FOLDER SLASH and (NOT OBJECT or IS NON-EMPTY OBJECT)
    # Scenarios for raising error:
    #   1. path is NOT OBJECT and NOT OBJECT PREFIX
    #   2. path is OBJECT and OBJECT PREFIX and HAS FOLDER SLASH and IS NON-EMPTY OBJECT
    if (path_is_object and not path_has_folder_slash) or (path_is_object and not path_is_prefix):
        logger.info(f"{s3_path} is an object. Downloading to {local_path} as file.")
        download_s3_object(
            s3_path=s3_path,
            local_path=local_path,
            exist_ok=exist_ok,
            transfer_config=transfer_config,
            force=force,
            size_only=size_only,
            **kwargs,
        )
    elif (path_is_prefix and not path_is_object) or (
        path_is_prefix and path_has_folder_slash and not path_is_non_empty_object_fn()
    ):
        logger.info(f"{s3_path} is an object prefix. Downloading to {local_path} as folder.")
        download_s3_object_prefix(
            s3_path=s3_path,
            local_path=local_path,
            exist_ok=exist_ok,
            transfer_config=transfer_config,
            force=force,
            size_only=size_only,
            **kwargs,
        )
    elif not path_is_object and not path_is_prefix:
        raise AWSError(f"{s3_path} is neither an object or prefix. Does it exist?")
    elif (
        path_is_object
        and path_is_prefix
        and path_has_folder_slash
        and path_is_non_empty_object_fn()
    ):
        raise AWSError(
            f"{s3_path} is an object and object prefix and has folder suffix "
            "and is non-empty. This is not supported"
        )
    else:
        raise AWSError(
            f"Unhandled scenario for {s3_path}: is_object={path_is_object}, "
            f"is_object_prefix={path_is_prefix}, "
            f"has_folder_slash={path_has_folder_slash}, "
            f"is_non_empty_object={path_is_non_empty_object_fn()}"
        )


def download_s3_object_prefix(
    s3_path: S3URI,
    local_path: Path,
    exist_ok: bool = False,
    transfer_config: Optional[TransferConfig] = None,
    force: bool = True,
    size_only: bool = False,
    **kwargs,
):
    """Download an S3 object prefix to a local path

    Args:
        s3_path (S3URI): URI of the object of folder
        local_path (Path): local destination path.
        exist_ok (bool, optional): If true, local path may exist previously. Defaults to False.

    """

    s3_object_paths = list_s3_paths(s3_path=s3_path, **kwargs)
    logger.info(f"Downloading {len(s3_object_paths)} objects under {local_path}")

    if not exist_ok and local_path.exists() and local_path.is_dir() and any(local_path.iterdir()):
        raise AWSError(
            f"{local_path} already exists and is not empty. Cannot download to the directory"
        )
    for s3_object_path in s3_object_paths:
        relative_key = s3_object_path.key[len(s3_path.key) :].lstrip("/")
        if s3_object_path.has_folder_suffix():
            if get_object(s3_object_path, **kwargs).content_length != 0:
                err_msg = (
                    f"Cannot download S3 object {s3_path} to {local_path}. Downloads of "
                    "objects with '/' suffixed keys and of non-zero size are NOT supported."
                )
                logger.error(err_msg)
                raise AWSError(err_msg)
            continue
        local_filepath = (local_path / relative_key).resolve()
        download_s3_object(
            s3_path=s3_object_path,
            local_path=local_filepath,
            exist_ok=exist_ok,
            transfer_config=transfer_config,
            force=force,
            size_only=size_only,
            **kwargs,
        )


@retry(
    (ConnectionClosedError, EndpointConnectionError, ResponseStreamingError, ClientError),
    tries=10,
    backoff=2.0,
)
def download_s3_object(
    s3_path: S3URI,
    local_path: Path,
    exist_ok: bool = False,
    transfer_config: Optional[TransferConfig] = None,
    force: bool = True,
    size_only: bool = False,
    **kwargs,
):
    """Download contents of an S3 object to file

    Args:
        s3_path (S3URI): S3 URI to object
        local_path (Path): destination path
        exist_ok (bool): If true, local path can already exist. Defaults to False.

    Raises:
        ValueError: Raised if local path exists.
    """
    s3_object = get_object(s3_path, **kwargs)
    if force or should_sync(
        source_path=s3_path, destination_path=local_path, size_only=size_only, **kwargs
    ):
        if local_path.exists() and not exist_ok and not force:
            raise ValueError(f"Unable to download S3 object to {local_path}. Path exists.")
        elif local_path.exists() and local_path.is_dir() and exist_ok:
            logger.warning(
                f"Overwriting directory {local_path} with S3 object {s3_path}"
                "This may cause unexpected behavior."
            )
            try:
                local_path.rmdir()
            except Exception as e:
                logger.error("Error removing directory: {e}")
                raise e
        local_path.parent.mkdir(parents=True, exist_ok=True)
        s3_object.download_file(Filename=str(local_path.resolve()), Config=transfer_config)


def upload_json(
    content: JSON, s3_path: S3URI, extra_args: Optional[Dict[str, Any]] = None, **kwargs
):
    with NamedTemporaryFile("w") as f:
        f.write(json.dumps(content, sort_keys=True))
        f.flush()

        upload_file(Path(f.name), s3_path=s3_path, extra_args=extra_args, **kwargs)


def upload_scratch_file(
    local_path: Path, s3_path: S3URI, extra_args: Optional[Dict[str, Any]] = None, **kwargs
):
    extra_args = extra_args or {}
    extra_args.update(SCRATCH_EXTRA_ARGS)
    upload_file(local_path=local_path, s3_path=s3_path, extra_args=extra_args, **kwargs)


def upload_path(
    local_path: Path,
    s3_path: S3URI,
    extra_args: Optional[Dict[str, Any]] = None,
    transfer_config: Optional[TransferConfig] = None,
    force: bool = True,
    size_only: bool = False,
    **kwargs,
):
    logger.info(f"Uploading contents at {local_path} to {s3_path}")
    if local_path.is_file():
        logger.info(f"{local_path} is a file.")
        upload_file(
            local_path=local_path,
            s3_path=s3_path,
            extra_args=extra_args,
            transfer_config=transfer_config,
            force=force,
            size_only=size_only,
            **kwargs,
        )
    elif local_path.is_dir():
        logger.info(f"{local_path} is a directory. Uploading all nested files to {s3_path}")
        upload_folder(
            local_path=local_path,
            s3_path=s3_path,
            extra_args=extra_args,
            transfer_config=transfer_config,
            force=force,
            size_only=size_only,
            **kwargs,
        )
    else:
        msg = f"Cannot upload {local_path} to S3. Path does not exist!"
        logger.error(msg)
        raise ValueError(msg)


def upload_folder(
    local_path: Path,
    s3_path: S3URI,
    extra_args: Optional[Dict[str, Any]] = None,
    transfer_config: Optional[TransferConfig] = None,
    force: bool = True,
    size_only: bool = False,
    **kwargs,
):
    local_paths = find_paths(local_path, include_dirs=False, include_files=True)
    for source_path in local_paths:
        destination_key = os.path.normpath(s3_path.key + source_path[len(str(local_path)) :])
        destination_path = S3URI.build(bucket_name=s3_path.bucket, key=destination_key)
        logger.debug(f"Uploading '{source_path}' to '{destination_path}'")
        upload_file(
            local_path=source_path,
            s3_path=destination_path,
            extra_args=extra_args,
            transfer_config=transfer_config,
            force=force,
            size_only=size_only,
            **kwargs,
        )
    logger.info(f"Uploaded {len(local_paths)} files to: {s3_path}")


@retry(ResponseStreamingError, tries=10, backoff=2.0)
def upload_file(
    local_path: Union[str, Path],
    s3_path: S3URI,
    extra_args: Optional[Dict[str, Any]] = None,
    transfer_config: Optional[TransferConfig] = None,
    force: bool = True,
    size_only: bool = False,
    **kwargs,
):
    s3_client = get_s3_client(**kwargs)
    if force or should_sync(
        source_path=Path(local_path), destination_path=s3_path, size_only=size_only, **kwargs
    ):
        s3_client.upload_file(
            Filename=str(local_path),
            Bucket=s3_path.bucket,
            Key=s3_path.key,
            ExtraArgs=extra_args or {},
            Config=transfer_config or TransferConfig(),
        )
    elif extra_args:
        # This handles scenario where extra args are specified but destination is more recent
        if "Tagging" in extra_args and "TaggingDirective" not in extra_args:
            extra_args["TaggingDirective"] = "REPLACE"
        if "Metadata" in extra_args and "MetadataDirective" not in extra_args:
            extra_args["MetadataDirective"] = "REPLACE"

        copy_s3_object(
            source_path=s3_path,
            destination_path=s3_path,
            extra_args=extra_args,
            transfer_config=transfer_config,
            force=True,
        )


def get_object(s3_path: S3URI, **kwargs) -> Object:
    s3 = get_s3_resource(**kwargs)
    try:
        return s3.Object(s3_path.bucket, s3_path.key)
    except ClientError as e:
        raise AWSError(f"Error finding s3 object: {s3_path} {e}") from e


def is_object(s3_path: S3URI, **kwargs) -> bool:
    s3 = get_s3_client(**kwargs)
    try:
        s3.head_object(Bucket=s3_path.bucket, Key=s3_path.key)
    except ClientError as e:
        if client_error_code_check(e, "404", "NoSuchKey", "NotFound"):
            return False
        raise AWSError(
            f"Error checking existence of {s3_path}: {get_client_error_message(e)}"
        ) from e
    return True


def is_object_prefix(s3_path: S3URI, **kwargs) -> bool:
    s3 = get_s3_client(**kwargs)
    response = s3.list_objects_v2(
        Bucket=s3_path.bucket,
        Prefix=s3_path.key,
        # StartAfter ensures we don't include objects at this key prefix.
        StartAfter=s3_path.key,
        MaxKeys=1,
    )
    return len(response.get("Contents", [])) > 0


def is_folder(s3_path: S3URI, **kwargs) -> bool:
    """Check if S3 Path is a "folder" or object

    To be a "folder", it must satisfy following conditions:
    - The Key Prefix HAS objects under prefix
    - All objects under Key Prefix are separated by '/'
        - i.e. key=key_prefix/...obj1, key=key_prefix/...obj2, etc.

    Example:

        For a bucket="s3://bucket" with the following keys:
            /path/to/one/object1
            /path/to/one/object2
            /path/to/one_object1
            /path/to/one_object2
            /path/to/another
            /path/to/another/object1
            /path/to/another/object2

        is_folder("s3://bucket/path/to/one") >> TRUE
        is_folder("s3://bucket/path/to/another") >> TRUE
        is_folder("s3://bucket/path/to/another/") >> TRUE

        is_folder("s3://bucket/path/to/one_") >> FALSE
        is_folder("s3://bucket/path/to/one_object1") >> FALSE
        is_folder("s3://bucket/path/to/doesnotexist") >> FALSE

    Args:
        s3_path (S3URI): S3 URI

    Returns:
        True if s3 path is a folder
    """
    return is_object_prefix(
        s3_path=S3URI.build(bucket_name=s3_path.bucket, key=s3_path.key_with_folder_suffix),
        **kwargs,
    )


def is_folder_placeholder_object(s3_path: S3URI, **kwargs) -> bool:
    """Check if S3 Path is a "folder placeholder" object

    A "folder placeholder" object is defined as an S3 object that:
    - Has a key that ends with a '/' character.
    - Has a content length of zero bytes.

    These objects are often used to represent folders in S3, which is a flat storage system.
    For these purposes, we want to ignore such objects when considering the contents of a folder.

    Args:
        s3_path (S3URI): S3 URI to check.

    Returns:
        bool: True if the S3 path is a folder placeholder object, False otherwise.
    """
    if not s3_path.has_folder_suffix():
        return False

    s3 = get_s3_client(**kwargs)
    try:
        obj = s3.head_object(Bucket=s3_path.bucket, Key=s3_path.key)
        return obj["ContentLength"] == 0
    except ClientError as e:
        if client_error_code_check(e, "404", "NoSuchKey", "NotFound"):
            return False
        raise AWSError(
            f"Error checking existence of {s3_path}: {get_client_error_message(e)}"
        ) from e


def get_s3_path_collection_stats(*s3_paths: S3URI, **kwargs) -> Mapping[S3URI, S3PathStats]:
    return dict(
        zip(
            s3_paths,
            parallel_starmap(get_s3_path_stats, [(_,) for _ in s3_paths], kwargs, ThreadPool),
        )
    )


def get_s3_path_stats(s3_path: S3URI, **kwargs) -> S3PathStats:
    """Adds some additional metadata to the file metadata to be stored
    in the GFS, such as ingestion time and file size.

    Args:
        s3_path (S3URI): Path to the file in s3.

    Returns:
        file_stats (dict): Dictionary object containing the additional
            values for the file stats.
    """
    s3 = get_s3_client(**kwargs)
    last_modified = get_current_time()
    bucket, key = s3_path.bucket, s3_path.key
    logger.info(f"Getting file stats: {s3_path}")

    size_bytes: Optional[int] = None
    object_count: Optional[int] = None

    try:
        file_info = s3.head_object(Bucket=bucket, Key=key)
        last_modified = file_info["LastModified"]
        size_bytes = file_info["ContentLength"]
        object_count = 1
    except ClientError:
        logger.debug("Caught client error, assuming it means this is a dir")
        # This means that the path is most likely a prefix, not an object key
        size_bytes, object_count = _get_prefix_size_and_count(
            bucket_name=bucket, key_prefix=key, **kwargs
        )
        last_modified = _get_prefix_last_modified(bucket_name=bucket, key_prefix=key, **kwargs)
    except Exception as e:
        logger.error("Caught unexpected exception.")
        logger.exception(e)
        raise e

    return S3PathStats(
        last_modified=last_modified, size_bytes=size_bytes, object_count=object_count
    )


# TODO: Two things
# 1. allow for a way to specify `size_only` for this fn when transfering large number of files
# 2. add flag for failing if no source data exists.
def sync_paths(
    source_path: Union[Path, S3URI],
    destination_path: Union[Path, S3URI],
    source_path_prefix: Optional[str] = None,
    include: Optional[List[Pattern]] = None,
    exclude: Optional[List[Pattern]] = None,
    extra_args: Optional[Dict[str, Any]] = None,
    transfer_config: Optional[TransferConfig] = None,
    force: bool = False,
    size_only: bool = False,
    delete: bool = False,
    **kwargs,
) -> List[S3TransferResponse]:
    logger.info(f"Syncing {source_path} to {destination_path}")

    source_path_key = source_path.key if isinstance(source_path, S3URI) else str(source_path)
    if not source_path_prefix:
        source_path_prefix = source_path_key
    elif not is_prefixed(source_path_key, source_path_prefix):
        raise ValueError(
            f"The source path prefix '{source_path_prefix}' does not match "
            f"source path {source_path}"
        )
    if isinstance(source_path, S3URI):
        nested_source_paths = list_s3_paths(
            s3_path=source_path.with_folder_suffix,
            include=include,
            exclude=exclude,
            **kwargs,
        )
        if is_object(source_path, **kwargs) and source_path not in nested_source_paths:
            nested_source_paths.insert(0, source_path)
    else:
        nested_source_paths = [
            Path(p)  # type: ignore
            for p in find_paths(
                root=source_path,
                include_dirs=False,
                includes=include,  # type: ignore
                excludes=exclude,  # type: ignore
            )
        ]

    logger.info(f"S3 Sync: Identified {len(nested_source_paths)} objects to sync")
    requests = [
        generate_transfer_request(
            source_path=nested_source_path,
            destination_path=destination_path,
            source_path_prefix=source_path_prefix,
            extra_args=extra_args,
        )
        for nested_source_path in nested_source_paths
    ]

    responses = process_transfer_requests(
        *requests,
        transfer_config=transfer_config,
        force=force,
        size_only=size_only,
        **kwargs,
    )

    if delete:
        logger.info("Sync: checking for files to delete following sync")
        if isinstance(destination_path, S3URI):
            unexpected_paths = set(list_s3_paths(destination_path, **kwargs)).difference(
                [S3URI(_.request.destination_path) for _ in responses]
            )
            logger.info(f"Sync: identified {len(unexpected_paths)} paths for deletion")
            for unexpected_path in unexpected_paths:
                delete_s3_path(unexpected_path, **kwargs)
        else:
            unexpected_local_paths = set(
                find_paths(destination_path, include_dirs=False)
            ).difference([str(_.request.destination_path) for _ in responses])
            logger.info(f"Sync: identified {len(unexpected_local_paths)} paths for deletion")
            for unexpected_local_path in unexpected_local_paths:
                remove_path(Path(unexpected_local_path))

    failures = 0
    for response in responses:
        failures += response.failed
    logger.info(f"Synced {len(requests) - failures} successfully. {failures} requests failed.")

    return responses


def generate_transfer_request(
    source_path: Union[Path, S3URI],
    destination_path: Union[Path, S3URI],
    source_path_prefix: Optional[str] = None,
    extra_args: Optional[Dict[str, Any]] = None,
) -> S3TransferRequest:
    """Create a S3 transfer request

    Args:
        source_path (Path|S3URI): source copy path
        destination_path (Path|S3URI): destination copy path
        source_path_prefix (str, optional): Optional prefix to remove from source path.
            Defaults to source path key.
    Returns:
        S3TransferRequest: _description_
    """
    relative_source_path = ""
    if source_path_prefix:
        source_key = source_path.key if isinstance(source_path, S3URI) else str(source_path)

        if not source_key.startswith(source_path_prefix):
            raise ValueError(
                f"Cannot generate S3CopyRequest with src={source_path}, dst={destination_path}. "
                f"source path prefix={source_path_prefix} is specified but does"
                " not match prefix or source path"
            )
        relative_source_path = source_key[len(source_path_prefix) :]

    if isinstance(destination_path, S3URI):
        # This will be sanitized by S3URI class (removing double slashes)
        new_destination_path = S3URI(destination_path + relative_source_path)
        if isinstance(source_path, S3URI):
            return S3CopyRequest(source_path, new_destination_path, extra_args=extra_args)
        else:
            return S3UploadRequest(source_path, new_destination_path, extra_args=extra_args)
    elif isinstance(source_path, S3URI) and isinstance(destination_path, Path):
        local_destination_path: Path = (
            Path(get_path_with_root(relative_source_path, destination_path))
            if relative_source_path
            else destination_path
        )
        return S3DownloadRequest(source_path, local_destination_path)
    else:
        raise ValueError("Local to local transfer is not ")


def process_transfer_requests(
    *transfer_requests: S3TransferRequest,
    transfer_config: Optional[TransferConfig] = None,
    force: bool = False,
    size_only: bool = False,
    suppress_errors: bool = False,
    **kwargs,
) -> List[S3TransferResponse]:
    """Process a list of S3 transfer requests

    Args:
        transfer_config (Optional[TransferConfig]): transfer config.
            Defaults to None.
        force (bool): Whether to force the transfer.
            Defaults to False.
        size_only (bool): Whether to only check size when transferring.
            Defaults to False.
        suppress_errors (bool): Whether to suppress errors.
            Defaults to False.

    Returns:
        List[S3TransferResponse]: List of transfer responses
    """
    transfer_responses = []

    for i, request in enumerate(transfer_requests):
        try:
            if isinstance(request, S3CopyRequest):
                copy_s3_object(
                    source_path=request.source_path,
                    destination_path=request.destination_path,
                    extra_args=request.extra_args,
                    transfer_config=transfer_config,
                    force=force,
                    size_only=size_only,
                    **kwargs,
                )
            elif isinstance(request, S3UploadRequest):
                upload_file(
                    local_path=request.source_path,
                    s3_path=request.destination_path,
                    extra_args=request.extra_args,
                    transfer_config=transfer_config,
                    force=force,
                    size_only=size_only,
                    **kwargs,
                )
            elif isinstance(request, S3DownloadRequest):
                # This is a special case where s3 object path has a trailing slash.
                # This should be ignored if the object is empty and raise an error otherwise.
                if request.source_path.has_folder_suffix():
                    if get_object(request.source_path, **kwargs).content_length != 0:
                        raise ValueError(
                            "Cannot download S3 object to local path. Downloads of objects "
                            "with '/' suffixed keys and of non-zero size are NOT supported."
                        )
                else:
                    download_s3_object(
                        s3_path=request.source_path,
                        local_path=request.destination_path,
                        transfer_config=transfer_config,
                        force=force,
                        size_only=size_only,
                        **kwargs,
                    )
            transfer_responses.append(S3TransferResponse(request, False))
            logger.info(f"Processed s3 transfer request {i + 1} of {len(transfer_requests)}")
        except Exception as e:
            msg = f"Failed to copy {request.source_path} to {request.destination_path}: {e}"
            if not suppress_errors:
                logger.error(msg)
                logger.exception(msg)
                raise e
            logger.warning(msg)
            transfer_responses.append(S3TransferResponse(request, True, f"{e}"))
    return transfer_responses


# Creating alias to enable easy recursive support
copy_s3_path = sync_paths


def client_error_code_check__SlowDown(ex: Exception) -> bool:
    return isinstance(ex, ClientError) and client_error_code_check(ex, "SlowDown")


@retry(ClientError, [client_error_code_check__SlowDown], tries=10, backoff=2.0)
def copy_s3_object(
    source_path: S3URI,
    destination_path: S3URI,
    extra_args: Optional[Dict[str, Any]] = None,
    transfer_config: Optional[TransferConfig] = None,
    force: bool = True,
    size_only: bool = False,
    **kwargs,
):
    s3 = get_s3_client(**kwargs)
    logger.info(f"S3: Copying {source_path} to {destination_path}")

    # This handles scenario where extra args are specified but destination is more recent
    if extra_args and "Tagging" in extra_args and "TaggingDirective" not in extra_args:
        extra_args["TaggingDirective"] = "REPLACE"
    if extra_args and "Metadata" in extra_args and "MetadataDirective" not in extra_args:
        extra_args["MetadataDirective"] = "REPLACE"

    if force or should_sync(
        source_path=source_path, destination_path=destination_path, size_only=size_only, **kwargs
    ):
        try:
            s3.copy(
                CopySource=source_path.as_dict(),
                Bucket=destination_path.bucket,
                Key=destination_path.key,
                ExtraArgs=extra_args or {},
                Config=transfer_config or TransferConfig(),
            )
        except ClientError as e:
            skippable_err_msg = (
                "This copy request is illegal because it is trying to copy an object "
                "to itself without changing the object's metadata, storage class, "
                "website redirect location or encryption attributes."
            )
            if get_client_error_message(e) != skippable_err_msg:
                raise e

    elif extra_args is not None:
        # This handles scenario where extra args are specified but destination is more recent
        copy_s3_object(
            source_path=destination_path,
            destination_path=destination_path,
            extra_args=extra_args,
            transfer_config=transfer_config,
            force=True,
            **kwargs,
        )


def delete_s3_path(
    s3_path: S3URI,
    include: Optional[List[Pattern]] = None,
    exclude: Optional[List[Pattern]] = None,
    **kwargs,
):
    """Deletes an S3 path (object or prefix)

    Args:
        s3_path (S3URI): Path or key prefix to delete
    """
    logger.info(f"Deleting S3 path {s3_path}")
    s3_paths = list_s3_paths(s3_path, include=include, exclude=exclude, **kwargs)
    delete_s3_objects(s3_paths, **kwargs)


def delete_s3_objects(s3_paths: List[S3URI], **kwargs):
    """Deletes a list of S3 objects

    Args:
        s3_paths (List[S3URI]): List of S3 paths to delete
    """
    logger.info(f"Found {len(s3_paths)} objects to delete.")
    s3 = get_s3_client(**kwargs)

    bucket_objects: Dict[str, Set[str]] = defaultdict(set)
    for s3_path in s3_paths:
        bucket_objects[s3_path.bucket].add(s3_path.key)

    # Can only specify a max of 1000 objects per request.
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.delete_objects
    MAX_KEYS_PER_REQUEST = 1000

    for bucket, keys in bucket_objects.items():
        key_list = list(keys)
        for i in range(0, len(keys), MAX_KEYS_PER_REQUEST):
            i_max = min(len(keys), i + MAX_KEYS_PER_REQUEST)
            logger.info(f"Deleting objects {i + 1}..{i_max + 1} in {bucket} bucket")
            s3.delete_objects(
                Bucket=bucket,
                Delete={"Objects": [{"Key": key_list[j]} for j in range(i, i_max)]},
            )


def move_s3_path(
    source_path: S3URI,
    destination_path: S3URI,
    include: Optional[List[Pattern]] = None,
    exclude: Optional[List[Pattern]] = None,
    extra_args: Optional[Dict[str, Any]] = None,
    transfer_config: Optional[TransferConfig] = None,
    **kwargs,
):
    """Move S3 Path from source to destination.

    There is no explicit "move" s3 method, so we combine COPY + DELETE operations

    Args:
        source_path (S3URI): source S3 Path
        destination_path (S3URI): destination S3 Path
        transfer_config (Optional[TransferConfig], optional): transfer config.
    """
    logger.info(f"Moving {source_path} to {destination_path}. Starting copy")
    responses = sync_paths(
        source_path=source_path,
        destination_path=destination_path,
        include=include,
        exclude=exclude,
        extra_args=extra_args,
        transfer_config=transfer_config,
        **kwargs,
    )
    logger.info(f"Copy complete. Starting deletion of {source_path}")
    paths_to_delete = [_.request.source_path for _ in responses if not _.failed]
    delete_s3_objects(paths_to_delete, **kwargs)


def list_s3_paths(
    s3_path: S3URI,
    include: Optional[List[Pattern]] = None,
    exclude: Optional[List[Pattern]] = None,
    **kwargs,
) -> List[S3URI]:
    """List all S3 paths under a Key prefix (as defined by S3 path)

    Include/Exclude patterns are applied to the RELATIVE KEY PATH

    Logic for how the include/exclude patterns are applied are as follows:

    - include/exclude: pattern provided? Y/N
    - I/E Match: If pattern provided, does s3 relative Key match? Y/N

    |  include | I Match | exclude | E Match | Append? |
    | ------------- No patterns provided ------------- |
    |     N    |    -    |    N    |    -    |    Y    |
    | ------ Include XOR Exclude pattern provided ---- |
    |     Y    |    Y    |    N    |    -    |    Y    |
    |     Y    |    N    |    N    |    -    |    N    |
    |     N    |    -    |    Y    |    Y    |    N    |
    |     N    |    -    |    Y    |    N    |    Y    |
    | ------ Include AND Exclude pattern provided ---- |
    |     Y    |    Y    |    Y    |    Y    |    N    |
    |     Y    |    N    |    Y    |    Y    |    N    |
    |     Y    |    Y    |    Y    |    N    |    Y    |
    |     Y    |    N    |    Y    |    N    |    N    |

    Args:
        s3_path (S3URI): The Root key path under which to find objects
        include (List[Pattern], optional): Optional list of regex patterns on which
            to retain objects if matching any. Defaults to None.
        exclude (List[Pattern], optional): Optional list of regex patterns on which
            to filter out objects if matching any. Defaults to None.

    Returns:
        List[S3URI]: List of S3 paths under root that satisfy filters
    """

    empty_include = (include is None) or (not any(include))
    empty_exclude = (exclude is None) or (not any(exclude))

    s3 = get_s3_client(**kwargs)

    def match_results(value: str, patterns: List[Pattern]) -> List[bool]:
        return [_.match(value) is not None for _ in patterns]

    paginator = s3.get_paginator("list_objects_v2")

    s3_paths: List[S3URI] = []
    for response in paginator.paginate(Bucket=s3_path.bucket, Prefix=s3_path.key):
        for item in response.get("Contents", []):
            key = item.get("Key", "")
            relative_key = key[len(s3_path.key) :]
            if empty_include or any(match_results(relative_key, include)):  # type: ignore
                if empty_exclude or (not any(match_results(relative_key, exclude))):  # type: ignore
                    s3_paths.append(S3URI.build(bucket_name=s3_path.bucket, key=key))
    return s3_paths


class PresignedUrlAction(Enum):
    READ = "get_object"
    WRITE = "put_object"


def generate_presigned_urls(
    s3_paths: List[S3URI],
    action: PresignedUrlAction = PresignedUrlAction.READ,
    expires_in: int = 3600,
    **kwargs,
) -> List[str]:
    """Generate Pre-signed URLs for given S3 Paths

    Args:
        s3_paths (List[S3URI]): List of S3 Paths to generate URLs for.
        action (PresignedUrlAction): Desired action for presigned URL (READ or WRITE),
            defaults to READ
        expires_in (int, optional): TTL of URL in seconds. Defaults to 3600.

    Returns:
        List[str]: List of pre-signed URLs
    """
    return [generate_presigned_url(s3_path, action, expires_in, **kwargs) for s3_path in s3_paths]


def generate_presigned_url(
    s3_path: S3URI,
    action: PresignedUrlAction = PresignedUrlAction.READ,
    expires_in: int = 3600,
    **kwargs,
) -> str:
    """Generate a Pre-signed URL for an S3 object

    Args:
        s3_path (S3URI): Intended S3 Path of the presigned URL
        action (PresignedUrlAction): Desired action for presigned URL (READ or WRITE),
            defaults to READ
        expires_in (int, optional): TTL of URL in seconds. Defaults to 3600.

    Returns:
        str: A Pre-signed URL
    """
    s3 = get_s3_client(config=Config(signature_version="s3v4"), **kwargs)
    presigned_url = s3.generate_presigned_url(
        ClientMethod=action.value,
        Params={"Bucket": s3_path.bucket, "Key": s3_path.key},
        HttpMethod="GET" if action == PresignedUrlAction.READ else "PUT",
        ExpiresIn=expires_in,
    )
    return presigned_url


def should_sync(
    source_path: Union[Path, S3URI],
    destination_path: Union[Path, S3URI],
    size_only: bool = False,
    **kwargs,
) -> bool:
    """Checks whether transfer from source to destination is required.

    This logic matches the logic in `aws s3 sync` command.

    A transfer from SRC -> DST is necessary if any of the following are true:
        - DST does not exist
        - SRC was last modified more recently than DST
        - SRC size is different than DST
        - not size_only and SRC ETag is differerent than DST

    Args:
        source_path (Union[Path, S3URI]): source path
        destination_path (Union[Path, S3URI]): destination to transfer to
        size_only (bool, optional): Limits content comparison to False, otherwise
    """
    source_last_modified: datetime
    source_size_bytes: int
    source_hash: Callable[[], Optional[str]]
    dest_last_modified: Optional[datetime] = None
    dest_size_bytes: Optional[int] = None
    dest_hash: Callable[[], Optional[str]]
    multipart_chunk_size_bytes: Optional[int] = None
    multipart_threshold_bytes: Optional[int] = None

    if isinstance(destination_path, S3URI) and is_object(destination_path):
        dest_s3_object = get_object(destination_path, **kwargs)
        dest_last_modified = dest_s3_object.last_modified
        dest_size_bytes = dest_s3_object.content_length
        multipart_chunk_size_bytes, multipart_threshold_bytes = determine_multipart_attributes(
            destination_path, **kwargs
        )

        def dest_hash() -> Optional[str]:
            return dest_s3_object.e_tag if not size_only else None
    elif isinstance(destination_path, Path) and destination_path.exists():
        dest_local_path = destination_path
        local_stats = dest_local_path.stat()
        dest_last_modified = datetime.fromtimestamp(local_stats.st_mtime, tz=timezone.utc)
        dest_size_bytes = local_stats.st_size

        def dest_hash() -> Optional[str]:
            return (
                get_local_etag(
                    dest_local_path, multipart_chunk_size_bytes, multipart_threshold_bytes
                )
                if not size_only
                else None
            )
    else:
        return True

    if isinstance(source_path, S3URI) and is_object(source_path):
        src_s3_object = get_object(source_path, **kwargs)
        source_last_modified = src_s3_object.last_modified
        source_size_bytes = src_s3_object.content_length
        multipart_chunk_size_bytes, multipart_threshold_bytes = determine_multipart_attributes(
            source_path, **kwargs
        )

        def source_hash() -> Optional[str]:
            return src_s3_object.e_tag if not size_only else None
    elif isinstance(source_path, Path) and source_path.exists():
        src_local_path = source_path
        local_stats = src_local_path.stat()
        source_last_modified = datetime.fromtimestamp(local_stats.st_mtime, tz=timezone.utc)
        source_size_bytes = local_stats.st_size

        def source_hash() -> Optional[str]:
            return (
                get_local_etag(
                    src_local_path, multipart_chunk_size_bytes, multipart_threshold_bytes
                )
                if not size_only
                else None
            )
    else:
        raise ValueError(
            f"Cannot transfer, source path {source_path} does not exist! "
            f"is s3={isinstance(source_path, S3URI)}, is local={isinstance(source_path, Path)} "
            f"is object={isinstance(source_path, S3URI) and is_object(source_path, **kwargs)}, "
            f"is local exists={isinstance(source_path, Path) and source_path.exists()}, "
            f"type={type(source_path)}"
        )

    if dest_size_bytes is None or dest_last_modified is None:
        return True
    if source_size_bytes != dest_size_bytes:
        return True
    if source_last_modified.replace(microsecond=0) > dest_last_modified.replace(microsecond=0):
        return True
    if not size_only and source_hash() != dest_hash():
        return True
    return False


def check_paths_in_sync(
    source_path: Union[Path, S3URI],
    destination_path: Union[Path, S3URI],
    size_only: bool = False,
    ignore_folder_placeholder_objects: bool = True,
    allow_subset: bool = False,
    max_workers: Optional[int] = None,
    **kwargs,
) -> bool:
    """Checks whether source and destination paths are in sync.

    Args:
        source_path (Union[Path, S3URI]): source path
        destination_path (Union[Path, S3URI]): destination path
        size_only (bool, optional): Limits content comparison to just size and date
            (no checksum/ETag). Defaults to False.
        ignore_folder_placeholder_objects (bool, optional): Whether to ignore S3 folder
            placeholder objects (zero-byte objects with keys ending in '/'). Defaults to True.
        allow_subset (bool, optional): Whether to allow source path to be a subset of
            destination path. Defaults to False.
        max_workers (Optional[int], optional): Number of worker threads to use for
            parallel comparison. Defaults to None (ThreadPoolExecutor default).

    Raises:
        ValueError: if the source path does not exist.

    Returns:
        bool: True if paths are in sync, False, otherwise
    """

    def _resolve_paths(path: Union[Path, S3URI]) -> List[Union[Path, S3URI]]:
        if isinstance(path, Path):
            if path.is_dir():
                return list(map(Path, sorted(find_paths(path, include_dirs=False))))
            else:
                return [path]
        else:
            if is_object(path, **kwargs) and not is_folder_placeholder_object(path, **kwargs):
                return [path]
            else:
                return [
                    _
                    for _ in sorted(list_s3_paths(path, **kwargs))
                    if (
                        not ignore_folder_placeholder_objects
                        or not is_folder_placeholder_object(_, **kwargs)
                    )
                ]

    def _find_relative_path(full_path: Union[Path, S3URI], root_path: Union[Path, S3URI]) -> str:
        if isinstance(full_path, Path) and isinstance(root_path, Path):
            if full_path == root_path:
                return ""
            # Adding the leading "/" to ensure we return the leading "/" in the relative path for
            # files under a folder. This is to be consistent with S3URI behavior.
            relative_path = "/" + strip_path_root(full_path, root_path)
            return relative_path
        elif isinstance(full_path, S3URI) and isinstance(root_path, S3URI):
            if full_path == root_path:
                return ""
            # Stripping the "/" to ensure we return the leading "/" in the relative path for
            # objects under a folder. This means that if we have:
            # root_path = `s3://bucket/folder/`
            # or root_path = `s3://bucket/folder`,
            # full_path = `s3://bucket/folder/subfolder/object.txt`
            # The relative path returned will be: `/subfolder/object.txt`

            relative_path = full_path.removeprefix(root_path.rstrip("/"))
            return relative_path
        else:
            raise ValueError("Mismatched path types between full_path and root_path")

    source_paths = _resolve_paths(source_path)
    destination_paths = _resolve_paths(destination_path)

    if len(source_paths) == 0:
        raise ValueError(f"Source path {source_path} does not exist")

    stripped_source_path_to_path_map: Dict[str, Union[Path, S3URI]] = {
        _find_relative_path(sp, source_path): sp for sp in source_paths
    }

    stripped_destination_path_to_path_map: Dict[str, Union[Path, S3URI]] = {
        _find_relative_path(dp, destination_path): dp for dp in destination_paths
    }

    missing_destination_paths = set(stripped_source_path_to_path_map.keys()).difference(
        stripped_destination_path_to_path_map.keys()
    )
    if missing_destination_paths:
        logger.info(
            "The following source paths are missing in the destination path: "
            f"{missing_destination_paths}"
        )
        return False
    if not allow_subset:
        extra_destination_paths = set(stripped_destination_path_to_path_map.keys()).difference(
            stripped_source_path_to_path_map.keys()
        )
        if extra_destination_paths:
            logger.info(
                "The following destination paths are extra compared to the source path: "
                f"{extra_destination_paths}"
            )
            return False

    # Run comparisons in parallel
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        future_to_pair = {
            executor.submit(
                should_sync,
                (sp := stripped_source_path_to_path_map[relative_path]),
                (dp := stripped_destination_path_to_path_map[relative_path]),
                size_only,
                **kwargs,
            ): (sp, dp)
            for relative_path in stripped_source_path_to_path_map
        }

        for future in as_completed(future_to_pair):
            not_ok = future.result()
            if not_ok:
                # Cancel any still-pending futures; we already know it's not in sync.
                for f in future_to_pair:
                    f.cancel()
                return False

    return True


def update_s3_storage_class(
    s3_path: S3URI,
    target_storage_class: S3StorageClass,
) -> bool:
    """Tries to transition an object (or objects) represented by an s3_path to a
    desired target_storage_class.

    NOTE: This function needs to be called again if it returns False.

    Args:
        s3_path (S3URI): The s3_path representing an s3 key or s3 prefix whose object or
            objects should have their storage class updated.
        target_storage_class (S3StorageClass): The target storage class.

    Raises:
        RuntimeError: If an unsupported target_storage_class is provided
        RuntimeError: If the path or paths under the provided root s3_path have a storage class
            that does not support transitions (e.g. S3StorageClass.OUTPOSTS,
            S3StorageClass.REDUCED_REDUNDANCY)

    Returns:
        bool:
            Returns True if the s3_path successfully had its storage class updated
            Returns False if s3_path did not fully update its storage class. Specifically:
            - A constituent object or objects under s3 archive storage class are still restoring
            - A constituent object or objects failed to transition to the desired storage class
    """

    if target_storage_class not in S3StorageClass.list_transitionable_storage_classes():
        raise RuntimeError(
            f"Error trying to update s3 storage class for s3_path ({s3_path}) "
            f"with unsupported target storage class ({target_storage_class.value})."
        )

    s3_paths = list_s3_paths(s3_path=s3_path)

    # 1. Iterate over all s3 paths under our s3_path and determine archive restores to be done.
    #    Also, start any storage class transitions that can be done.
    paths_to_restore: List[S3URI] = []
    paths_restoring: List[S3URI] = []
    failed_transitions: List[S3URI] = []
    for p in s3_paths:
        run_path_transition: bool = False
        s3_obj = get_object(p)
        print(
            f"debug: current storage class: {s3_obj.storage_class}, target: {target_storage_class}"
        )
        current_storage_class = S3StorageClass.from_boto_s3_obj(s3_obj)  # type: ignore[arg-type]
        # Current storage class matches target: No-op
        if current_storage_class == target_storage_class:
            continue
        # Current storage class is archived: Check restore status
        elif current_storage_class in S3StorageClass.list_archive_storage_classes():
            o = S3RestoreStatus.from_raw_s3_restore_status(s3_obj.restore)
            print(
                f"s3 path ({p}), current: {current_storage_class}, "
                f"target: {target_storage_class}, restore status: {o}"
            )
            if o.restore_status == S3RestoreStatusEnum.NOT_STARTED:
                paths_to_restore.append(p)
            elif o.restore_status == S3RestoreStatusEnum.IN_PROGRESS:
                paths_restoring.append(p)
            elif o.restore_status == S3RestoreStatusEnum.FINISHED:
                run_path_transition = True
        # Current storage class does not match target: Needs transition
        elif current_storage_class in S3StorageClass.list_transitionable_storage_classes():
            run_path_transition = True
        # Current storage class cannot be transitioned
        else:
            raise RuntimeError(
                f"Error trying to update the s3 storage class for s3_path ({p}) "
                f"which has an unsupported current storage class: {current_storage_class}"
            )

        if run_path_transition:
            try:
                copy_s3_object(
                    source_path=p,
                    destination_path=p,
                    extra_args={"StorageClass": target_storage_class.value},
                )
            except ClientError as e:
                logger.error(
                    f"Failed to transition s3 path ({p}) to storage "
                    f"class ({target_storage_class.value}). Error details: {e}"
                )
                failed_transitions.append(p)

    # 2. Start off any restores that need to be done
    for p in paths_to_restore:
        s3_obj = get_object(p)
        s3_obj.restore_object(
            RestoreRequest={
                "Days": 2,
                "GlacierJobParameters": {"Tier": "Standard"},
            }
        )

    # 3. If there are restores (started or in progress) or failed transitions for objects
    #    under our `s3_path` return False and we'll need to call update_s3_storage class in
    #    the future again.
    if any(paths_to_restore) or any(paths_restoring):
        return False

    if any(failed_transitions):
        logger.warning(
            "The following paths failed to transition to the target_storage_class "
            f"({target_storage_class}): {failed_transitions}"
        )
        return False

    return True


# --------------------------------------------------------------------
#                               Helpers
# --------------------------------------------------------------------
def _get_prefix_size_and_count(bucket_name: str, key_prefix: str, **kwargs) -> Tuple[int, int]:
    s3 = get_s3_resource(**kwargs)
    bucket = s3.Bucket(bucket_name)
    object_count = 0
    total_size = 0
    for obj in bucket.objects.filter(Prefix=key_prefix):
        total_size += obj.size
        object_count += 1
    return total_size, object_count


def _get_prefix_last_modified(bucket_name: str, key_prefix: str, **kwargs) -> datetime:
    s3 = get_s3_resource(**kwargs)
    bucket = s3.Bucket(bucket_name)
    last_modified = None
    for obj in bucket.objects.filter(Prefix=key_prefix):
        if not last_modified or obj.last_modified > last_modified:
            last_modified = obj.last_modified
    assert isinstance(last_modified, datetime)  # mollify mypy
    return last_modified


def determine_multipart_attributes(
    s3_path: S3URI, **kwargs
) -> Tuple[Optional[int], Optional[int]]:
    """Determines multipart upload chunk size and approximate threshold, if applicable

    Multipart attributes are the following:
        - chunk size: The size of each part in bytes
        - threshold: The size in bytes at which a multipart upload is created

    Notes:
        - The threshold is only an approximation, as it is not possible to
          determine the exact threshold used.
        - This assumes chunk size is constant for all parts.
          This is not necessarily always true, although rare.
        - The chunksize for a multipart upload has the following constraints:
            https://docs.aws.amazon.com/AmazonS3/latest/userguide/qfacts.html
            - Most importantly, must be between 5MB and 5GB
    Args:
        s3_path (S3URI): S3 object to check

    Returns:
        Tuple[Optional[int], Optional[int]]: (chunk_size, threshold)
    """
    s3_client = get_s3_client(**kwargs)
    head_object_part = s3_client.head_object(Bucket=s3_path.bucket, Key=s3_path.key, PartNumber=1)

    object_parts = head_object_part.get("PartsCount", 1)
    object_part_content_length = head_object_part.get("ContentLength", 0)

    threshold, chunk_size = None, None
    if object_parts > 1:
        threshold, chunk_size = object_part_content_length, object_part_content_length
    else:
        head_object = s3_client.head_object(Bucket=s3_path.bucket, Key=s3_path.key)
        is_multipart_etag = head_object["ETag"].endswith('-1"')
        if is_multipart_etag:
            # This should ensure that multipart etag is created as expected
            threshold = object_part_content_length
        if object_part_content_length > AWS_S3_DEFAULT_CHUNK_SIZE_BYTES:
            chunk_size = object_part_content_length
            if not is_multipart_etag:
                chunk_size += 1
                threshold = chunk_size
    return chunk_size, threshold


def determine_chunk_size(
    path: Path, default_chunk_size_bytes: int = AWS_S3_DEFAULT_CHUNK_SIZE_BYTES
) -> int:
    """Function to determine the chunk size that `aws s3 cp` would use for a very large file"""
    file_size = path.stat().st_size
    correct_chunk_size_bytes = default_chunk_size_bytes
    while math.ceil(file_size / correct_chunk_size_bytes) > AWS_S3_MULTIPART_LIMIT:
        correct_chunk_size_bytes *= 2
    return correct_chunk_size_bytes


@retry(OSError)
def get_local_etag(
    path: Path, chunk_size_bytes: Optional[int] = None, threshold_bytes: Optional[int] = None
) -> str:
    """Calculates an expected AWS s3 upload etag for a local on-disk file.
    Takes into account multipart uploads, but does NOT account for additional encryption
    (like KMS keys)

    Args:
        path (Path): The path of the file that will be uploaded to s3
        chunk_size (int): The default multipart upload chunksize in bytes.
            If None, we determine the chunk size based on file size

    Returns:
        str: The expected etag
    """
    if chunk_size_bytes is None:
        chunk_size_bytes = determine_chunk_size(path)

    if threshold_bytes is None:
        threshold_bytes = determine_chunk_size(path)

    size_bytes = path.stat().st_size

    chunk_digests: List[bytes] = []
    buffer_size = min(chunk_size_bytes, LOCAL_ETAG_READ_BUFFER_BYTES)

    with open(path, "rb") as fp:
        chunk_hasher = hashlib.md5()
        chunk_bytes_read = 0

        while True:
            if chunk_bytes_read == chunk_size_bytes:
                chunk_digests.append(chunk_hasher.digest())
                chunk_hasher = hashlib.md5()
                chunk_bytes_read = 0

            bytes_remaining_in_chunk = chunk_size_bytes - chunk_bytes_read
            read_size = min(buffer_size, bytes_remaining_in_chunk)
            data = fp.read(read_size)
            if not data:
                break

            chunk_hasher.update(data)
            chunk_bytes_read += len(data)

        if chunk_bytes_read:
            chunk_digests.append(chunk_hasher.digest())

    if len(chunk_digests) == 0:  # Empty file (can never be multipart upload)
        expected_etag = f'"{hashlib.md5().hexdigest()}"'
    elif len(chunk_digests) == 1 and size_bytes < threshold_bytes:  # File smaller than threshold
        expected_etag = f'"{chunk_digests[0].hex()}"'
    else:  # We are dealing with a multipart upload
        digests = b"".join(chunk_digests)
        multipart_md5 = hashlib.md5(digests)
        expected_etag = f'"{multipart_md5.hexdigest()}-{len(chunk_digests)}"'

    return expected_etag



__all__ = [
    "get_sns_client",
    "get_sns_resource",
    "publish_to_topic",
    "publish",
]

import logging
from typing import TYPE_CHECKING, Mapping, Optional

from botocore.exceptions import ClientError

from aibs_informatics_aws_utils.core import AWSService
from aibs_informatics_aws_utils.exceptions import AWSError

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_sns.type_defs import (
        MessageAttributeValueTypeDef,
        PublishInputTypeDef,
        PublishResponseTypeDef,
    )
else:
    PublishResponseTypeDef = dict
    MessageAttributeValueTypeDef = dict
    PublishInputTypeDef = dict


logger = logging.getLogger(__name__)


get_sns_client = AWSService.SNS.get_client
get_sns_resource = AWSService.SNS.get_resource


def publish(
    message: str,
    topic_arn: Optional[str] = None,
    target_arn: Optional[str] = None,
    phone_number: Optional[str] = None,
    subject: Optional[str] = None,
    message_structure: Optional[str] = None,
    message_attributes: Optional[Mapping[str, MessageAttributeValueTypeDef]] = None,
    message_deduplication_id: Optional[str] = None,
    message_group_id: Optional[str] = None,
    region: Optional[str] = None,
) -> PublishResponseTypeDef:
    if topic_arn is None and target_arn is None and phone_number is None:
        raise AWSError("Must provide either a topic_arn, target_arn, or phone_number")
    sns = get_sns_client(region=region)
    logger.info(
        f"Publishing message: {message} with subject: {subject}, "
        f"to topic: {topic_arn}, target: {target_arn}, phone_number: {phone_number}"
    )
    request: PublishInputTypeDef = PublishInputTypeDef(Message=message)
    if topic_arn:
        request["TopicArn"] = topic_arn
    if target_arn:
        request["TargetArn"] = target_arn
    if phone_number:
        request["PhoneNumber"] = phone_number
    if message:
        request["Message"] = message
    if subject:
        request["Subject"] = subject
    if message_structure:
        request["MessageStructure"] = message_structure
    if message_attributes:
        request["MessageAttributes"] = message_attributes
    if message_deduplication_id:
        request["MessageDeduplicationId"] = message_deduplication_id
    if message_group_id:
        request["MessageGroupId"] = message_group_id
    try:
        publish_response = sns.publish(**request)
    except ClientError as e:
        logger.exception(e)
        raise AWSError(f"Could not publish message using request parameters: {request}")
    return publish_response


def publish_to_topic(
    message: str,
    topic_arn: str,
    subject: Optional[str] = None,
    message_structure: Optional[str] = None,
    message_attributes: Optional[Mapping[str, MessageAttributeValueTypeDef]] = None,
    message_deduplication_id: Optional[str] = None,
    message_group_id: Optional[str] = None,
    region: Optional[str] = None,
) -> PublishResponseTypeDef:
    return publish(
        message=message,
        topic_arn=topic_arn,
        subject=subject,
        message_structure=message_structure,
        message_attributes=message_attributes,
        message_deduplication_id=message_deduplication_id,
        message_group_id=message_group_id,
        region=region,
    )



from __future__ import annotations

__all__ = ["BaseFileSystem", "LocalFileSystem", "S3FileSystem"]

import errno
import os
from abc import abstractmethod
from collections import deque
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Union

import pytz
from aibs_informatics_core.models.aws.efs import EFSPath
from aibs_informatics_core.models.aws.s3 import S3URI
from aibs_informatics_core.models.base import CustomAwareDateTime, custom_field
from aibs_informatics_core.models.base.model import SchemaModel
from aibs_informatics_core.utils.logging import get_logger
from aibs_informatics_core.utils.os_operations import find_all_paths
from aibs_informatics_core.utils.time import BEGINNING_OF_TIME
from aibs_informatics_core.utils.tools.strtools import removeprefix

from aibs_informatics_aws_utils.efs import get_efs_path, get_local_path
from aibs_informatics_aws_utils.s3 import get_s3_resource

logger = get_logger(__name__)

SEP = "/"


@dataclass
class PathStats(SchemaModel):
    size_bytes: int = custom_field()
    object_count: int = custom_field()
    last_modified: datetime = custom_field(mm_field=CustomAwareDateTime())


@dataclass(order=True)
class Node:
    """Represents an object or folder in an file system path.

    Args:
        path_part (str): specifies the key part of the fs path (an edge) to this node
        parent (Optional[Node]): Optionally specify the parent node to which
            this node is connected. By default, this is None.
        children (Dict[str, Node]): Child nodes that exist under this path prefix.
        size_bytes (int): The size (in bytes) of all objects under this path prefix.
        object_count (int): The number of objects under this path prefix.
        last_modified (datetime): The most recent date any objects under this prefix were
            last modified.

    """

    path_part: str
    parent: Optional["Node"] = field(default=None)
    children: Dict[str, "Node"] = field(default_factory=dict)
    size_bytes: int = field(default=0)
    object_count: int = field(default=0)
    last_modified: datetime = field(default=BEGINNING_OF_TIME)
    is_path_part_prefix: bool = field(default=False)
    is_path_part_suffix: bool = field(default=False)

    def __hash__(self) -> int:
        return hash(self.path)

    @property
    def key(self) -> str:
        return self.path

    @property
    def parent_path(self) -> str:
        parent_path = self.parent.path if self.parent else ""
        if parent_path and self.is_path_part_suffix:
            parent_path = parent_path.rstrip(SEP)
        return parent_path

    @property
    def normalized_path_part(self) -> str:
        path_part = self.path_part
        if self.has_children() and not self.is_path_part_prefix:
            path_part = path_part.rstrip(SEP) + SEP
        return path_part

    @property
    def path(self) -> str:
        return self.parent_path + self.normalized_path_part

    @property
    def path_stats(self) -> PathStats:
        return PathStats(
            size_bytes=self.size_bytes,
            object_count=self.object_count,
            last_modified=self.last_modified,
        )

    @property
    def depth(self) -> int:
        return self.parent.depth + 1 if self.parent else 0

    def has_children(self) -> bool:
        return not (len(self.children) < 1)

    def add_object(self, path: str, size: int, last_modified: datetime):
        def _add_object(node: Node, path: Optional[str]):
            node._update_stats(size=size, last_modified=last_modified)
            if path is None:
                return

            first_key_part, remaining_key = path.split(SEP, 1) if SEP in path else (path, None)
            if first_key_part:
                if first_key_part not in node.children:
                    node.children[first_key_part] = Node(path_part=first_key_part, parent=node)
                node = node.children[first_key_part]
            _add_object(node, remaining_key)

        # TODO: Right now, we cannot support non-folder prefixes
        _add_object(self, path.lstrip(SEP))

    def get(self, key: str) -> Optional["Node"]:
        try:
            return self[key]
        except KeyError:
            return None

    def list_nodes(self) -> List["Node"]:
        nodes = [self]
        for _, n in self.children.items():
            nodes.extend(n.list_nodes())
        return nodes

    def _update_stats(self, size: int, last_modified: datetime):
        # For each node, update the current node's stats
        self.size_bytes += size
        self.object_count += 1
        if self.last_modified < last_modified:
            self.last_modified = last_modified

    def __getitem__(self, key: str) -> "Node":
        _self = self
        for key_part in key.split(SEP):
            # Only access if value is not empty string
            if key_part:
                _self = _self.children[key_part]
        return _self

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"path_part={self.path_part}, "
            f"children={len(self.children)}, "
            f"size_bytes={self.size_bytes}, "
            f"object_count={self.object_count})"
        )


@dataclass  # type: ignore[misc] # mypy #5374
class BaseFileSystem:
    node: Node = field(init=False)

    def __post_init__(self):
        self.node = self.initialize_node()

    @abstractmethod
    def initialize_node(self) -> Node:
        raise NotImplementedError()

    @abstractmethod
    def refresh(self, **kwargs):
        raise NotImplementedError()

    def partition(
        self,
        size_bytes_limit: Optional[int] = None,
        object_count_limit: Optional[int] = None,
        raise_error_if_criteria_not_met: bool = False,
    ) -> List[Node]:
        """Partitions the root tree folder structure into a list of nodes

        Partitioning is guided by constraints by size and object count.

        Args:
            size_bytes_limit (Optional[int], optional): If specified, partitions must be
                less than the specified value.
            object_count_limit (Optional[int], optional): If specified, partitions
                must contain fewer objects than the specified valude
            raise_error_if_criteria_not_met (bool, optional): If True, raises error if nodes
                cannot meet criteria. In actuality, this is more relevant for size limitations
                where an object size is greater than the size limit

        Raises:
            ValueError: Thrown if raise_error_if_criteria_not_met is true and criteria not met

        Returns:
            List[Node]: List of nodes partitioning
        """
        unchecked_nodes = {self.node}
        size_bytes_exceeding_obj_nodes = []

        partitioned_nodes: List[Node] = []
        logger.info(
            f"Partitioning nodes with size_bytes_limit={size_bytes_limit} "
            f"and object_count_limit={object_count_limit}"
        )

        while unchecked_nodes:
            unchecked_node = unchecked_nodes.pop()
            if (size_bytes_limit and unchecked_node.size_bytes > size_bytes_limit) or (
                object_count_limit and unchecked_node.object_count > object_count_limit
            ):
                if unchecked_node.has_children():
                    unchecked_nodes.update(unchecked_node.children.values())
                else:
                    size_bytes_exceeding_obj_nodes.append(unchecked_node)
            else:
                partitioned_nodes.append(unchecked_node)

        if size_bytes_exceeding_obj_nodes:
            msg = (
                f"Found {len(size_bytes_exceeding_obj_nodes)} objects that exceed the "
                f"partition size limit {size_bytes_limit}."
            )
            if raise_error_if_criteria_not_met:
                raise ValueError(msg)
            logger.warning(msg)
            partitioned_nodes.extend(size_bytes_exceeding_obj_nodes)
        logger.info(f"Partitioned {len(partitioned_nodes)} nodes.")
        return partitioned_nodes

    @classmethod
    @abstractmethod
    def from_path(cls, path: str, **kwargs) -> BaseFileSystem:
        pass


@dataclass
class LocalFileSystem(BaseFileSystem):
    path: Path

    def initialize_node(self) -> Node:
        return Node(path_part=self.path.as_posix())

    def refresh(self, **kwargs):
        self.node = self.initialize_node()
        paths_to_visit = deque(find_all_paths(self.path, include_dirs=False, include_files=True))
        while paths_to_visit:
            path = paths_to_visit.popleft()
            try:
                path_stats = Path(path).stat()
                self.node.add_object(
                    path=removeprefix(path, str(self.path) + os.sep),
                    size=path_stats.st_size,
                    last_modified=datetime.fromtimestamp(path_stats.st_mtime, tz=pytz.UTC),
                )
            except FileNotFoundError:
                logger.warning(f"{path} does not exist. Not adding to {self}")
            except OSError as ose:
                # Suppress error if Stale File. This is expected error if file has been deleted:
                #   - https://stackoverflow.com/a/40351967
                #   - https://www.rfc-editor.org/rfc/rfc7530#section-4
                if ose.errno == errno.ESTALE:
                    logger.warning(f"{ose} raised for {path}.")
                    if Path(path).exists():
                        logger.warning(f"Adding {path} to end of list to check later.")
                        paths_to_visit.append(path)
                else:
                    logger.error(f"Unexpected error raised for {path}. Reason: {ose}")
                    raise ose

    @classmethod
    def from_path(cls, path: Union[str, Path], **kwargs) -> LocalFileSystem:
        local_path = Path(path)
        local_root = LocalFileSystem(path=local_path)
        local_root.refresh(**kwargs)
        return local_root


@dataclass
class EFSFileSystem(LocalFileSystem):
    efs_path: EFSPath

    def initialize_node(self) -> Node:
        return Node(path_part=self.efs_path)

    @classmethod
    def from_path(cls, path: Union[str, Path], **kwargs) -> EFSFileSystem:
        if isinstance(path, str) and EFSPath.is_valid(path):
            efs_path = EFSPath(path)
            local_path = get_local_path(efs_path=efs_path)
        else:
            local_path = Path(path)
            efs_path = get_efs_path(local_path=local_path)

        efs_root = EFSFileSystem(path=local_path, efs_path=efs_path)
        efs_root.refresh(**kwargs)
        return efs_root


@dataclass
class S3FileSystem(BaseFileSystem):
    """Generates a FS tree structure of an S3 path with size and object count stats

    Args:
        bucket (str): The S3 bucket to describe
        key (str): The S3 key to describe
        node (Node): The underlying root node of FS tree
    """

    bucket: str
    key: str

    def initialize_node(self) -> Node:
        return Node(path_part=self.key)

    def refresh(self, **kwargs):
        self.node = self.initialize_node()
        s3 = get_s3_resource(**kwargs)
        bucket = s3.Bucket(self.bucket)

        for obj in bucket.objects.filter(Prefix=self.key):
            self.node.add_object(
                path=removeprefix(obj.key, self.key),
                size=obj.size,
                last_modified=obj.last_modified,
            )

    @classmethod
    def from_path(cls, path: str, **kwargs) -> S3FileSystem:
        s3_path = S3URI(path)
        s3_root = S3FileSystem(bucket=s3_path.bucket, key=s3_path.key)
        s3_root.refresh(**kwargs)
        return s3_root


def get_file_system(path: Union[str, Path]) -> BaseFileSystem:
    if isinstance(path, str) and S3URI.is_valid(path):
        return S3FileSystem.from_path(path)
    elif isinstance(path, str) and EFSPath.is_valid(path):
        return EFSFileSystem.from_path(path)
    else:
        return LocalFileSystem.from_path(path)



import logging
import mimetypes
from email.mime.multipart import MIMEMultipart
from email.mime.nonmultipart import MIMENonMultipart
from email.mime.text import MIMEText
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, Union

from aibs_informatics_core.models.email_address import EmailAddress
from botocore.exceptions import ClientError

from aibs_informatics_aws_utils.core import AWSService, get_region
from aibs_informatics_aws_utils.exceptions import AWSError

if TYPE_CHECKING:
    # TypeDefs can be used as kwargs:
    # Example 1: SendEmailRequestTypeDef = {}, ses.send_email(**kwargs)
    # Example 2: SendRawEmailRequestTypeDef = {}, ses.send_raw_email(**kwargs)
    from mypy_boto3_ses.type_defs import (
        DestinationTypeDef,
        MessageTagTypeDef,
        MessageTypeDef,
        SendEmailRequestTypeDef,
        SendEmailResponseTypeDef,
        SendRawEmailRequestTypeDef,
        SendRawEmailResponseTypeDef,
    )

    # 'Request' portion of name for RequestRequestTypeDefs is not accidentally repeated
    # See: https://youtype.github.io/boto3_stubs_docs/mypy_boto3_ses/type_defs/#sendrawemailrequestrequesttypedef
else:
    (
        SendEmailRequestTypeDef,
        SendEmailResponseTypeDef,
        SendRawEmailRequestTypeDef,
        SendRawEmailResponseTypeDef,
        DestinationTypeDef,
        MessageTagTypeDef,
        MessageTypeDef,
    ) = (dict, dict, dict, dict, dict, dict, dict)


logger = logging.getLogger(__name__)


get_ses_client = AWSService.SES.get_client


def verify_email_identity(email_address: str) -> Dict[str, Any]:  # no type def?
    """
    TODO: replace with send_custom_verification_email-> SendCustomVerificationEmailResponseTypeDef:
    """
    ses = get_ses_client(region=get_region())
    response = ses.verify_email_identity(EmailAddress=email_address)
    return response


def is_verified(identity: str) -> bool:
    """Checks if email address or domain identity is valid

    Examples:
        email identitiy:    `is_verified(myemail@subdomain.domain.com)`
        subdomain identity  `is_verified(subdomain.domain.com)`
        domain idenitity    `is_verified(domain.com)`

    """
    ses = get_ses_client(region=get_region())
    try:
        response = ses.get_identity_verification_attributes(Identities=[identity])
        v_attrs = response["VerificationAttributes"]
        if identity not in v_attrs:
            return False
        return v_attrs[identity]["VerificationStatus"] == "Success"
    except ClientError as e:
        logger.exception(e)
        raise AWSError(f"Could not check verification status, error: {e}")


def send_email(request: SendEmailRequestTypeDef) -> SendEmailResponseTypeDef:
    logger.info(f"Sending email request: {request}")
    ses = get_ses_client(region=get_region())

    try:
        response = ses.send_email(**request)
    except ClientError as e:
        logger.exception(e.response)
        raise AWSError(f"Could not send email, error: {e}, {e.response}")

    return response


def send_simple_email(
    source: Union[str, EmailAddress],
    to_addresses: Sequence[Union[str, EmailAddress]],
    subject: str,
    body: str = "",
) -> SendEmailResponseTypeDef:
    return send_email(
        SendEmailRequestTypeDef(
            Source=source,
            Destination={"ToAddresses": to_addresses},
            Message={"Subject": {"Data": subject}, "Body": {"Text": {"Data": body}}},
        )
    )


def send_raw_email(request: SendRawEmailRequestTypeDef) -> SendRawEmailResponseTypeDef:
    logger.info(f"Sending email request: {request}")
    ses = get_ses_client(region=get_region())

    try:
        response = ses.send_raw_email(**request)
    except ClientError as e:
        logger.exception(e.response)
        raise AWSError(f"Could not send email, error: {e}, {e.response}")

    return response


def send_email_with_attachment(
    source: Union[str, EmailAddress],
    to_addresses: Sequence[Union[str, EmailAddress]],
    subject: str,
    body: Union[str, MIMEText] = "",
    attachments_paths: Optional[List[Path]] = None,
) -> SendRawEmailResponseTypeDef:
    """
    Args:
        source: Source email address
        to_addresses: List of recipient email addresses
        subject: Email subject
        body: Email body which can be either basic str or MIMEText, which can
            allow html with hyperlinks.
        attachments_paths: List of optional paths to read contents from and attach to the email
    Returns: `SendEmailResponseTypeDef`
    """
    msg = MIMEMultipart("mixed")
    msg["Subject"] = subject
    msg["From"] = source
    msg["To"] = ", ".join(to_addresses)

    msg_body = MIMEMultipart("alternative")
    if isinstance(body, str):
        msg_body.attach(MIMEText(body))
    else:
        msg_body.attach(body)
    msg.attach(msg_body)

    if attachments_paths is not None:
        for attachments_path in attachments_paths:
            attachment_obj = _construct_mime_attachment_from_path(path=attachments_path)
            msg.attach(attachment_obj)

    return send_raw_email(SendRawEmailRequestTypeDef(RawMessage={"Data": msg.as_string()}))


def _construct_mime_attachment_from_path(path: Path) -> MIMENonMultipart:
    """Constructs a MIME attachment from a `Path`"""
    mimetype, _ = mimetypes.guess_type(url=path)

    if mimetype is None:
        raise RuntimeError(f"Could not guess the MIME type for the file/object at: {path}")

    maintype, subtype = mimetype.split("/")

    filename = Path(path).name

    with open(path) as f:
        data = f.read()

    mime_obj = MIMENonMultipart(maintype, subtype)
    mime_obj.set_payload(data)
    mime_obj["Content-Type"] = mimetype
    mime_obj.add_header("Content-Disposition", "attachment", filename=filename)

    return mime_obj



import json
from typing import Any, Dict, Literal, Optional, overload

from aibs_informatics_aws_utils.core import AWSService

get_secretsmanager_client = AWSService.SECRETSMANAGER.get_client


@overload
def get_secret_value(
    secret_name: str, as_dict: Literal[False] = False, region: Optional[str] = None
) -> str: ...


@overload
def get_secret_value(
    secret_name: str, as_dict: Literal[True], region: Optional[str] = None
) -> Dict[str, Any]: ...


def get_secret_value(secret_name: str, as_dict: bool = False, region: Optional[str] = None):
    """Retrieves a Secrets Manager secret value

    Args:
        secret_name (str): the Secrets Manager secret name

    Raises:
        ValueError: If there is no such key

    Returns:
        str: The secret value stored at the key name
    """
    secretsmanager = get_secretsmanager_client(region=region)

    response = secretsmanager.get_secret_value(SecretId=secret_name)

    secret = response["SecretString"]
    if as_dict:
        return json.loads(secret)
    else:
        return secret



# --------------------------------------------------------------------------
# DynamoDB Exceptions
# --------------------------------------------------------------------------


class DBException(Exception):
    """Generic Exception for any DynamoDB related Errors"""


class DBWriteException(DBException):
    """Generic Exception for DynamoDB *Put/Update/Delete* related Errors"""


class DBReadException(DBException):
    """Generic Exception for DynamoDB *Read Invocation* related Errors"""


class DBQueryException(DBReadException):
    """Generic Exception for DynamoDB *Query Invocation* related Errors"""


class DBQueryResultException(DBReadException):
    """Generic Exception for DynamoDB *Query Result* related Errors"""


class NonUniqueQueryResultException(DBQueryResultException):
    """Raised when 1 query result is expected but more are fewer are returned"""


class EmptyQueryResultException(DBQueryResultException):
    """Raised when at least 1 query result was expected but none were returned"""


# --------------------------------------------------------------------------
# Other AWS Exceptions
# --------------------------------------------------------------------------


class AWSError(ValueError):
    """AWS related Exception"""


class InvalidAmazonResourceNameError(AWSError):
    """Raised when ARN is invalid"""


class ResourceNotFoundError(AWSError):
    """Raised when AWS resource not found"""



__all__ = [
    "ECRImage",
    "ECRImageUri",
    "ECRLogin",
    "ECRRegistry",
    "ECRRegistryUri",
    "ECRRepository",
    "ECRRepositoryUri",
    "ECRResource",
    "LifecyclePolicy",
    "LifecyclePolicyAction",
    "LifecyclePolicyRule",
    "ResourceTag",
    "TagMode",
    "get_ecr_client",
    "resolve_image_uri",
    "ECRImageReplicator",
    "ReplicateImageRequest",
    "ReplicateImageResponse",
]

from .core import (
    ECRImage,
    ECRImageUri,
    ECRLogin,
    ECRRegistry,
    ECRRegistryUri,
    ECRRepository,
    ECRRepositoryUri,
    ECRResource,
    LifecyclePolicy,
    LifecyclePolicyAction,
    LifecyclePolicyRule,
    ResourceTag,
    TagMode,
    get_ecr_client,
    resolve_image_uri,
)
from .image_replicator import ECRImageReplicator, ReplicateImageRequest, ReplicateImageResponse



import base64
import json
import logging
import re
from copy import deepcopy
from dataclasses import dataclass, field
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Type,
    TypedDict,
    TypeVar,
    Union,
    cast,
)

import requests
from aibs_informatics_core.collections import StrEnum, ValidatedStr
from aibs_informatics_core.models.base import DataClassModel
from aibs_informatics_core.utils.logging import LoggingMixin
from botocore.exceptions import ClientError
from dataclasses_json import LetterCase, config

from aibs_informatics_aws_utils.core import (
    AWSService,
    get_account_id,
    get_client_error_code,
    get_region,
)
from aibs_informatics_aws_utils.exceptions import (
    AWSError,
    InvalidAmazonResourceNameError,
    ResourceNotFoundError,
)

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_ecr import ECRClient
    from mypy_boto3_ecr.literals import ImageTagMutabilityType, TagStatusType
    from mypy_boto3_ecr.type_defs import (
        BatchGetImageResponseTypeDef,
        ImageDetailTypeDef,
        ImageIdentifierTypeDef,
        LayerTypeDef,
        ListImagesFilterTypeDef,
        ListImagesRequestPaginateTypeDef,
        TagTypeDef,
    )
else:
    ECRClient = object
    ImageTagMutabilityType = object
    TagStatusType = object
    BatchGetImageResponseTypeDef = dict
    ImageDetailTypeDef = dict
    ImageIdentifierTypeDef = dict
    LayerTypeDef = dict
    ListImagesFilterTypeDef = dict
    ListImagesRequestPaginateTypeDef = dict
    TagTypeDef = dict

logger = logging.getLogger(__name__)


get_ecr_client = AWSService.ECR.get_client

ECR_REPO_ARN_PATTERN = re.compile(
    r"arn:aws:ecr:([\w-]*):([\d]{10,12}):repository/((?:[a-z0-9]+(?:[._-][a-z0-9]+)*/)*[a-z0-9]+(?:[._-][a-z0-9]+)*)"
)

ECR_REGISTRY_PATTERN_STR = r"([\d]{12}).dkr.ecr.([\w-]*).amazonaws.com"
ECR_REPO_PATTERN_STR = r"/((?:[a-z0-9]+(?:[._-][a-z0-9]+)*/)*[a-z0-9]+(?:[._-][a-z0-9]+)*)"
ECR_IMAGE_PATTERN_STR = r"(:([\w.\-_]{1,127})|@(sha256:[A-Fa-f0-9]{64}))"

ECR_REGISTRY_URI_PATTERN = re.compile(ECR_REGISTRY_PATTERN_STR)
ECR_REPO_URI_PATTERN = re.compile(ECR_REGISTRY_PATTERN_STR + ECR_REPO_PATTERN_STR)
ECR_IMAGE_URI_PATTERN = re.compile(
    ECR_REGISTRY_PATTERN_STR + ECR_REPO_PATTERN_STR + ECR_IMAGE_PATTERN_STR
)


class ECRRegistryUri(ValidatedStr):
    regex_pattern = ECR_REGISTRY_URI_PATTERN

    @property
    def account_id(self) -> str:
        return self.get_match_groups()[0]

    @property
    def region(self) -> str:
        return self.get_match_groups()[1]

    @classmethod
    def from_components(
        cls, account_id: Optional[str] = None, region: Optional[str] = None, **kwargs
    ) -> "ECRRegistryUri":
        """Generate a Registry URI.

        If account ID not provided, account Id of credentials is used.
        If region is not provided, region of credentials is used.

        Args:
            account_id (str, optional): The registry ID. Defaults to None.
            region (str, optional): AWS region. Defaults to None.

        Returns:
            str: AWS Registry URI
        """
        account_id = account_id or get_account_id()
        region = get_region(region)
        return ECRRegistryUri(f"{account_id}.dkr.ecr.{region}.amazonaws.com")


class ECRRepositoryUri(ECRRegistryUri):
    regex_pattern = ECR_REPO_URI_PATTERN

    @property
    def repository_name(self) -> str:
        return self.get_match_groups()[2]

    @classmethod
    def from_components(  # type: ignore[override]
        cls,
        repository_name: str,
        account_id: Optional[str] = None,
        region: Optional[str] = None,
    ) -> "ECRRepositoryUri":
        """Generate a Repository URI.

        If account ID not provided, account Id of credentials is used.
        If region is not provided, region of credentials is used.

        Args:
            repository_name (str): Name of ECR repository
            account_id (str, optional): The registry ID. Defaults to None.
            region (str, optional): AWS region. Defaults to None.

        Returns:
            str: AWS Repository URI
        """
        registry_uri = ECRRegistryUri.from_components(account_id=account_id, region=region)
        return ECRRepositoryUri(f"{registry_uri}/{repository_name}")


class ECRImageUri(ECRRepositoryUri):
    regex_pattern = ECR_IMAGE_URI_PATTERN

    @property
    def image_tag(self) -> Optional[str]:
        return self.get_match_groups()[-2]

    @property
    def image_digest(self) -> Optional[str]:
        return self.get_match_groups()[-1]

    @classmethod
    def from_components(  # type: ignore[override]
        cls,
        repository_name: str,
        image_tag: Optional[str] = None,
        image_digest: Optional[str] = None,
        account_id: Optional[str] = None,
        region: Optional[str] = None,
    ) -> "ECRImageUri":
        """Generate a Image URI.

        If account ID not provided, account Id of credentials is used.
        If region is not provided, region of credentials is used.

        Args:
            repository_name (str): Name of ECR repository
            image_tag (str, optional): tag associated. Defaults to None.
            image_digest (str, optional): the image digest. Defaults to None.
            account_id (str, optional): The registry ID. Defaults to None.
            region (str, optional): AWS region. Defaults to None.

        Raises:
            ValueError: If both or neither image tag / image digest are provided.

        Returns:
            str: ECR Image URI
        """
        if (image_tag and image_digest) or (not image_tag and not image_digest):
            raise ValueError(
                "Must provide EITHER image tag OR image digest. "
                f"image_tag={image_tag}, image_digest={image_digest}"
            )
        repo_uri = ECRRepositoryUri.from_components(
            repository_name=repository_name, account_id=account_id, region=region
        )
        image_id = f"{'@' if image_digest else ':'}{image_digest or image_tag}"
        return ECRImageUri(f"{repo_uri}{image_id}")


@dataclass
class ECRLogin:
    username: str
    password: str
    registry: str

    @property
    def auth_token(self) -> str:
        decoded_auth_token = f"{self.username}:{self.password}"
        return base64.b64encode(decoded_auth_token.encode()).decode()


@dataclass
class LifecyclePolicySelection(DataClassModel):
    tag_status: Literal["tagged", "untagged", "any"] = field(
        metadata=config(letter_case=LetterCase.CAMEL)
    )
    tag_prefix_list: List[str] = field(metadata=config(letter_case=LetterCase.CAMEL))
    count_type: Literal["imageCountMoreThan", "sinceImagePushed"] = field(
        metadata=config(letter_case=LetterCase.CAMEL)
    )
    count_number: int = field(metadata=config(letter_case=LetterCase.CAMEL))
    count_unit: Optional[Literal["days"]] = field(
        default=None, metadata=config(letter_case=LetterCase.CAMEL)
    )

    def __post_init__(self):
        if self.count_type != "sinceImagePushed" and self.count_unit is not None:
            # https://docs.aws.amazon.com/AmazonECR/latest/userguide/LifecyclePolicies.html#lp_count_unit
            raise ValueError(f"Cannot specify 'countUnit' when countType={self.count_type}")
        if self.tag_status == "tagged" and not self.tag_prefix_list:
            # https://docs.aws.amazon.com/AmazonECR/latest/userguide/LifecyclePolicies.html#lp_tag_prefix_list
            raise ValueError(f"Must specify 'tagPrefixList' when tagStatus={self.tag_status}")


@dataclass
class LifecyclePolicyAction(DataClassModel):
    type: Literal["expire"] = field(
        default="expire", metadata=config(letter_case=LetterCase.CAMEL)
    )


@dataclass
class LifecyclePolicyRule(DataClassModel):
    rule_priority: int = field(metadata=config(letter_case=LetterCase.CAMEL))
    description: str = field(metadata=config(letter_case=LetterCase.CAMEL))
    selection: LifecyclePolicySelection = field(metadata=config(letter_case=LetterCase.CAMEL))
    action: LifecyclePolicyAction = field(
        default_factory=LifecyclePolicyAction,
        metadata=config(letter_case=LetterCase.CAMEL),
    )

    @classmethod
    def REMOVE_UNTAGGED(cls, rule_priority: int = 1, days: int = 14) -> "LifecyclePolicyRule":
        return LifecyclePolicyRule(
            rule_priority=rule_priority,
            description=f"Remove untagged images after {days} days",
            selection=LifecyclePolicySelection(
                tag_status="untagged",
                tag_prefix_list=None,  # type: ignore[arg-type]
                count_type="sinceImagePushed",
                count_number=days,
                count_unit="days",
            ),
            action=LifecyclePolicyAction("expire"),
        )


@dataclass
class LifecyclePolicy(DataClassModel):
    rules: List[LifecyclePolicyRule]

    def __post_init__(self):
        self.rules = self.reprioritize_rules(self.rules, in_place=True)

    @staticmethod
    def reprioritize_rules(
        rules: List[LifecyclePolicyRule], in_place: bool = False
    ) -> List[LifecyclePolicyRule]:
        rule_priority = 0

        sorted_rules: List[LifecyclePolicyRule]

        def sort_key(_):
            return _.rule_priority

        if in_place:
            rules.sort(key=sort_key)
            sorted_rules = rules
        else:
            sorted_rules = sorted(deepcopy(rules), key=sort_key)

        for rule in sorted_rules:
            if rule.rule_priority <= rule_priority:
                rule_priority += 1
                rule.rule_priority = rule_priority
            else:
                rule_priority = rule.rule_priority
        return sorted_rules

    @classmethod
    def from_rules(cls, *rules: LifecyclePolicyRule) -> "LifecyclePolicy":
        return LifecyclePolicy(list(map(deepcopy, rules)))


class ResourceTag(TypedDict):
    Key: str
    Value: str


class TagMode(StrEnum):
    OVERWRITE = "overwrite"
    APPEND = "append"


T = TypeVar("T", bound="ECRMixins")


@dataclass
class ECRMixins(LoggingMixin):
    account_id: str
    region: str

    def __init__(self, account_id: str, region: str, client: Optional[ECRClient] = None):
        self.account_id = account_id
        self.region = region
        self._client = client

    @property
    def client(self) -> ECRClient:
        if self._client is None:
            self._client = get_ecr_client(self.region)
        return self._client

    @client.setter
    def client(self, value: ECRClient):
        self._client = value  # pragma: no cover

    @property
    def uri(self) -> str:
        raise NotImplementedError("")  # pragma: no cover

    @classmethod
    def from_uri(cls: Type[T], uri: str) -> T:
        raise NotImplementedError("")  # pragma: no cover


@dataclass
class ECRImage(ECRMixins, DataClassModel):
    repository_name: str
    image_digest: str
    # https://distribution.github.io/distribution/spec/manifest-v2-2/#image-manifest-field-descriptions
    image_manifest: str = field(default=None, repr=False)  # type: ignore[assignment]

    def __init__(
        self,
        account_id: str,
        region: str,
        repository_name: str,
        image_digest: str,
        image_manifest: Optional[str] = None,
        client: Optional[ECRClient] = None,
    ):
        super().__init__(account_id=account_id, region=region, client=client)
        self.repository_name = repository_name
        self.image_digest = image_digest
        if image_manifest is None:
            response = self.client.batch_get_image(
                repositoryName=self.repository_name,
                registryId=self.account_id,
                imageIds=[ImageIdentifierTypeDef(imageDigest=self.image_digest)],
            )
            if len(response["images"]) == 0 or "imageManifest" not in response["images"][0]:
                raise ResourceNotFoundError(f"Could not resolve image manifest for {self.uri}")

            self.image_manifest = response["images"][0]["imageManifest"]
        else:
            self.image_manifest = image_manifest

    @property
    def image_pushed_at(self) -> Optional[datetime]:
        return self.get_image_detail().get("imagePushedAt")

    @property
    def image_tags(self) -> List[str]:
        image_detail = self.get_image_detail()
        return image_detail.get("imageTags", [])

    @property
    def uri(self) -> str:
        return ECRImageUri.from_components(
            repository_name=self.repository_name,
            image_digest=self.image_digest,
            account_id=self.account_id,
            region=self.region,
        )

    def get_repository(self) -> "ECRRepository":
        return ECRRepository(
            account_id=self.account_id,
            region=self.region,
            repository_name=self.repository_name,
        )

    def get_image_detail(self) -> ImageDetailTypeDef:
        """Get image detail of this image from ECR

        Returns:
            ImageDetailTypeDef: image detail dictionary
        """
        response = self.client.describe_images(
            repositoryName=self.repository_name,
            registryId=self.account_id,
            imageIds=[ImageIdentifierTypeDef(imageDigest=self.image_digest)],
        )
        image_details = response["imageDetails"]
        if len(image_details) == 0:
            raise ResourceNotFoundError(
                f"Could not resolve image detail for {self}"
            )  # pragma: no cover
        return image_details[0]

    def get_image_layers(self) -> List[LayerTypeDef]:
        """Get layers from image manifest into ECR Layer objects

        The schema of the image manifest layers is defined here:
        https://distribution.github.io/distribution/spec/manifest-v2-2/#image-manifest-field-descriptions

        Note: While docker image manifests can have multiple formats, ECR only supports
              the schema defined in the link above, a v2 single image manifest. There is
              a manifest list, that describes multiple architectures, but ECR does not support
              this. This method assumes the image manifest is in the correct format.

        Returns:
            List[LayerTypeDef]: List of ECR Image layers
        """
        image_manifest = json.loads(self.image_manifest)

        return [
            LayerTypeDef(
                layerDigest=layer["digest"],
                layerAvailability="AVAILABLE",
                layerSize=layer["size"],
                mediaType=layer["mediaType"],
            )
            for layer in image_manifest["layers"]
        ]

    def get_image_config_layer(self) -> LayerTypeDef:
        """Get the image config layer from image manifest

        The schema of the image manifest config layer is defined here:
        https://distribution.github.io/distribution/spec/manifest-v2-2/#image-manifest-field-descriptions

        Note: While docker image manifests can have multiple formats, ECR only supports
              the schema defined in the link above, a v2 single image manifest. There is
              a manifest list, that describes multiple architectures, but ECR does not support
              this. This method assumes the image manifest is in the correct format.

        Returns:
            List[LayerTypeDef]: Layer Type dict of the config object
        """
        image_manifest = json.loads(self.image_manifest)
        layer = image_manifest["config"]
        return LayerTypeDef(
            layerDigest=layer["digest"],
            layerAvailability="AVAILABLE",
            layerSize=layer["size"],
            mediaType=layer["mediaType"],
        )

    def get_image_config(self) -> Dict[str, Any]:
        """Get ECR or docker image configuration json metadata
        Returns:
            Dict[str, Any]: dictionary with image configuration
        """
        # get image_manifest (sha256: hash)
        config_digest = json.loads(self.image_manifest)["config"]["digest"]
        registry = ECRRegistry(account_id=self.account_id, region=self.region)
        ecr_login = registry.get_ecr_login()

        response = requests.get(
            url=f"https://{registry.uri}/v2/{self.repository_name}/blobs/{config_digest}",
            headers={"Authorization": f"Basic {ecr_login.auth_token}"},
        )
        response.raise_for_status()
        return response.json()

    def add_image_tags(self, *image_tags: str):
        """
        Add tags to image

        Args:
            image_tags (Union[str, None]): list of tags to add to image
        """
        self.logger.info(f"Adding tags={image_tags} to {self.uri}")
        for tag in image_tags:
            self.put_image(image_tag=tag)

    def put_image(self, image_tag: Optional[str]):
        """Make a call to put_image API to add image to ECR repository

        This method will add an image to the ECR repository. If the image already exists,
        it will not raise an error. Instead, it will log that the image already
        exists with the given tag.

        Note: This operation does not push an image to the repository. It only adds
        the image manifest to the repository.

        Args:
            image_tag (Optional[str]): tag to associate with image. If None, image is untagged.

        """
        try:
            if image_tag is None:
                self.client.put_image(
                    registryId=self.account_id,
                    repositoryName=self.repository_name,
                    imageManifest=self.image_manifest,
                    imageDigest=self.image_digest,
                )
            else:
                self.client.put_image(
                    registryId=self.account_id,
                    repositoryName=self.repository_name,
                    imageManifest=self.image_manifest,
                    imageDigest=self.image_digest,
                    imageTag=image_tag,
                )
        except ClientError as e:
            # IF we receive a ImageAlreadyExistsException,
            # then we have nothing to worry about. Otherwise, raise error.
            if get_client_error_code(e) != "ImageAlreadyExistsException":
                raise e
            self.logger.info(f"Image already exists with tag={image_tag}")
        else:
            self.logger.info(f"Added new image with tag={image_tag}")

    @classmethod
    def from_repository_uri(cls, repository_uri: str, image_digest: str) -> "ECRImage":
        repo_uri = ECRRepositoryUri(repository_uri)
        return ECRImage.from_uri(f"{repo_uri}@{image_digest}")

    @classmethod
    def from_uri(cls, uri: str) -> "ECRImage":
        image_uri = ECRImageUri(uri)

        account_id = image_uri.account_id
        region = image_uri.region
        repo_name = image_uri.repository_name
        image_tag = image_uri.image_tag
        image_digest = image_uri.image_digest

        ecr = get_ecr_client(region)
        if image_digest is None:
            image_details = ecr.describe_images(
                repositoryName=repo_name,
                registryId=account_id,
                imageIds=[ImageIdentifierTypeDef(imageTag=image_tag)] if image_tag else [],
            )["imageDetails"][0]
            assert "imageDigest" in image_details
            image_digest = image_details["imageDigest"]

        return ECRImage(
            account_id=account_id,
            region=region,
            repository_name=repo_name,
            image_digest=image_digest,
        )

    def __repr__(self) -> str:
        return (
            f"ECRImage("
            f"account_id='{self.account_id}', "
            f"region='{self.region}', "
            f"repository_name='{self.repository_name}', "
            f"image_digest={self.image_digest[7:15]}..{self.image_digest[-4:]})"
        )


class ECRResource(ECRMixins, DataClassModel):
    @property
    def arn(self) -> str:
        return f"arn:aws:ecr:{self.region}:{self.account_id}"

    def get_resource_tags(self) -> List[ResourceTag]:
        """Gets the tags for this ECR Resource

        Returns:
            List[ResourceTag]: list of tags
        """
        return [
            ResourceTag(Key=tag["Key"], Value=tag["Value"])
            for tag in self.client.list_tags_for_resource(resourceArn=self.arn)["tags"]
            if "Key" in tag and "Value" in tag
        ]

    def update_resource_tags(
        self, *tags: ResourceTag, mode: TagMode = cast(TagMode, TagMode.APPEND)
    ):
        """Updates the tags for a ECR Resource

        An update can either append or overwrite the existing tags.


        Args:
            mode (TagMode, optional): Either append or overwrite tags of resource.
                Defaults to TagMode.APPEND.
        """
        tag_dict = dict([(tag["Key"], tag["Value"]) for tag in tags])
        if mode == TagMode.OVERWRITE:
            existing_tags = self.get_resource_tags()
            tag_keys_to_remove = [
                existing_tag["Key"]
                for existing_tag in existing_tags
                if existing_tag["Key"] not in tag_dict
            ]
            self.client.untag_resource(resourceArn=self.arn, tagKeys=tag_keys_to_remove)
        self.client.tag_resource(
            resourceArn=self.arn,
            tags=[TagTypeDef(Key=key, Value=value) for key, value in tag_dict.items()],
        )


@dataclass
class ECRRepository(ECRResource):
    repository_name: str

    def __init__(
        self,
        account_id: str,
        region: str,
        repository_name: str,
        client: Optional[ECRClient] = None,
    ):
        super().__init__(account_id=account_id, region=region, client=client)
        self.repository_name = repository_name

    @property
    def arn(self) -> str:
        return f"{super().arn}:repository/{self.repository_name}"

    @property
    def uri(self) -> str:
        return ECRRepositoryUri.from_components(
            repository_name=self.repository_name,
            account_id=self.account_id,
            region=self.region,
        )

    def create(
        self,
        tags: Optional[List[ResourceTag]] = None,
        image_tag_mutability: ImageTagMutabilityType = "MUTABLE",
        exists_ok: bool = True,
    ):
        """Create an ECR Repository

        Args:
            tags (List[ResourceTag], optional): list of repo tags to add. Defaults to None.
            image_tag_mutability (ImageTagMutabilityType, optional): whether image
                tag is immutable. Defaults to "MUTABLE".
            exists_ok (bool, optional): Suppress error if repository already exists.
                Defaults to True.

        """
        if tags is None:
            tags = []

        try:
            self.client.create_repository(
                registryId=self.account_id,
                repositoryName=self.repository_name,
                tags=tags,
                imageTagMutability=image_tag_mutability,
            )
        except ClientError as e:
            # If the repo already exists, just move on.
            if get_client_error_code(e) == "RepositoryAlreadyExistsException" and exists_ok:
                # update tags
                if tags:
                    self.update_resource_tags(*tags)
            else:
                raise e

    def exists(self) -> bool:
        """Check if repository exists

        Returns:
            bool: True if repository exists
        """
        try:
            self.client.describe_repositories(
                registryId=self.account_id, repositoryNames=[self.repository_name]
            )
        except ClientError as e:
            if get_client_error_code(e) == "RepositoryNotFoundException":
                return False
            else:
                raise e
        else:
            return True

    def put_lifecycle_policy(self, lifecycle_policy: LifecyclePolicy):
        self.client.put_lifecycle_policy(
            repositoryName=self.repository_name,
            lifecyclePolicyText=lifecycle_policy.to_json(sort_keys=True),
            registryId=self.account_id,
        )

    def delete(self, force: bool):
        """Delete the ECR repository described by this instance

        Args:
            force (bool): Ignore if images in repository.
        """
        self.client.delete_repository(
            registryId=self.account_id, repositoryName=self.repository_name, force=force
        )

    def get_image(
        self, image_tag: Optional[str] = None, image_digest: Optional[str] = None
    ) -> "ECRImage":
        """Get the image associated with the following tag.

        Args:
            image_tag (str): image tag

        Returns:
            ECRImage: the image with the image tag
        """
        if (image_tag is None) == (image_digest is None):
            raise ValueError(
                f"Must provide image tag XOR digest. "
                f"provided image_tag={image_tag}, image_digest={image_digest}"
            )
        images = self.get_images()
        for image in images:
            if image_tag and image_tag in image.image_tags:
                return image
            elif image_digest and image_digest == image.image_digest:
                return image
        else:
            raise ResourceNotFoundError(
                f"Could not find an image in {self.uri} with tag={image_tag}"
            )

    def get_images(self, tag_status: TagStatusType = "ANY") -> List["ECRImage"]:
        """Fetches all images in a given repository

        Args:
            tag_status (TagStatusType, optional): Filter non-tagged images. Defaults to "ANY".

        Returns:
            List[ECRImage]: list of images in repository
        """

        # To fetch necessary image info, we need to make two calls:
        #   1. Call the `list_images` API which will return imageDigest/imageTag.
        #       - This is a lightweight response which gives us digests
        #   2. Call the `batch_get_image` API which returns imageManifest info.
        #       - this is consolidated all in one go

        # Call 1: list_images
        paginator = self.client.get_paginator("list_images")
        list_image_request = ListImagesRequestPaginateTypeDef(
            repositoryName=self.repository_name,
            filter=ListImagesFilterTypeDef(tagStatus=tag_status),
        )
        image_digests: List[str] = sorted(
            list(
                {
                    image_id["imageDigest"]
                    for list_images_response in paginator.paginate(**list_image_request)
                    for image_id in list_images_response["imageIds"]
                    if "imageDigest" in image_id
                }
            )
        )
        if len(image_digests) == 0:
            return []

        # Call 2: batch_get_image
        response: BatchGetImageResponseTypeDef = self.client.batch_get_image(
            repositoryName=self.repository_name,
            registryId=self.account_id,
            imageIds=[ImageIdentifierTypeDef(imageDigest=digest) for digest in image_digests],
        )

        # Next we consolidate the results, ensuring that the image manifests
        # are all the same. If an image digest has differing manifests,
        # we should throw an error.
        digest_to_manifest_map: Dict[str, str] = {}
        for image in response["images"]:
            image_digest = image["imageId"]["imageDigest"]  # type: ignore
            image_manifest = image["imageManifest"]  # type: ignore
            if image_digest in digest_to_manifest_map:
                if image_manifest != digest_to_manifest_map[image_digest]:
                    raise ValueError(
                        f"Not all image manifests are equivalent for {image_digest} in {self.uri}"
                    )
            else:
                digest_to_manifest_map[image_digest] = image_manifest

        return [
            ECRImage(
                account_id=self.account_id,
                region=self.region,
                repository_name=self.repository_name,
                image_digest=image_digest,
                image_manifest=image_manifest,
            )
            for image_digest, image_manifest in digest_to_manifest_map.items()
        ]

    @classmethod
    def from_uri(cls, uri: str) -> "ECRRepository":
        repo_uri = ECRRepositoryUri(uri)
        return ECRRepository(
            account_id=repo_uri.account_id,
            region=repo_uri.region,
            repository_name=repo_uri.repository_name,
        )

    @classmethod
    def from_arn(cls, arn: str) -> "ECRRepository":
        match = ECR_REPO_ARN_PATTERN.match(arn)
        if not match:
            raise InvalidAmazonResourceNameError(
                f"resource ARN = '{arn}' does not match {ECR_REPO_ARN_PATTERN} pattern"
            )
        return ECRRepository(
            account_id=match.group(2),
            region=match.group(1),
            repository_name=match.group(3),
        )

    @classmethod
    def from_name(
        cls,
        repository_name: str,
        account_id: Optional[str] = None,
        region: Optional[str] = None,
    ) -> "ECRRepository":
        region = get_region(region)
        account_id = account_id or get_account_id()
        return ECRRepository(account_id=account_id, region=region, repository_name=repository_name)


class ECRRegistry(ECRResource):
    @property
    def uri(self) -> str:
        return ECRRegistryUri.from_components(account_id=self.account_id, region=self.region)

    def get_repositories(
        self,
        repository_name: Optional[Union[str, re.Pattern]] = None,
        repository_tags: Optional[List[ResourceTag]] = None,
    ) -> List[ECRRepository]:
        """Filter repositories based on resource tags specified

        Args:
            repositories (List[ECRRepository]): list of resources to filter
            filter_tags (List[ResourceTag]): list of resource tags

        Returns:
            List[ECRRepository]: filtered list with resource tags
        """
        repositories = self.list_repositories()
        filtered_repos = []
        for repo in repositories:
            if repository_tags:
                repo_tags = repo.get_resource_tags()
                if not all([filter_tag in repo_tags for filter_tag in repository_tags]):
                    continue
            if repository_name:
                if isinstance(repository_name, re.Pattern):
                    if not repository_name.match(repo.repository_name):
                        continue
                elif repository_name not in repo.repository_name:
                    continue
            filtered_repos.append(repo)
        return filtered_repos

    def list_repositories(self) -> List[ECRRepository]:
        """List all repositories in the Registry

        Returns:
            List[ECRRepository]: list of repositories in the registry
        """

        paginator = self.client.get_paginator("describe_repositories")
        repositories: List[ECRRepository] = []
        for describe_repos_response in paginator.paginate(registryId=self.account_id):
            for repository in describe_repos_response["repositories"]:
                assert "repositoryName" in repository
                repositories.append(
                    ECRRepository(
                        account_id=self.account_id,
                        region=self.region,
                        repository_name=repository["repositoryName"],
                    )
                )
        return repositories

    def get_ecr_login(self) -> ECRLogin:
        auth = self.client.get_authorization_token(registryIds=[self.account_id])
        if len(auth["authorizationData"]) != 1:
            raise AWSError(f"Could not resolve authorization token. Reponse: {auth}")
        auth_data_entry = auth["authorizationData"][0]
        auth_token = base64.b64decode(auth_data_entry["authorizationToken"]).decode()  # type: ignore
        username, password = auth_token.split(":")
        registry = auth_data_entry["proxyEndpoint"].replace("https://", "")  # type: ignore
        self.logger.debug(f"Registry (proxy endpoint): {registry}")
        assert "expiresAt" in auth_data_entry
        expires_at = auth_data_entry["expiresAt"].isoformat()
        self.logger.debug(f"Token expires at: {expires_at}")

        return ECRLogin(username=username, password=password, registry=registry)

    @classmethod
    def from_uri(cls, uri: str) -> "ECRRegistry":
        registry_uri = ECRRegistryUri(uri)
        return ECRRegistry(account_id=registry_uri.account_id, region=registry_uri.region)

    @classmethod
    def from_env(cls, region: Optional[str] = None) -> "ECRRegistry":
        return ECRRegistry(account_id=get_account_id(), region=get_region(region))


def resolve_image_uri(name: str, default_tag: Optional[str] = None) -> str:
    """Resolve full image URI from input name

    Args:
        name (str): Partial or fully qualified uri, name of image or repository

    Returns:
        str: fully qualified image URI
    """

    try:
        uri = name

        if not ECRRegistryUri.is_prefixed(uri):
            uri = f"{ECRRegistryUri.from_components()}/{uri}"

        if ECRImageUri.is_valid(uri):
            return uri
        elif ECRRepositoryUri.is_valid(uri):
            repo = ECRRepository.from_uri(uri)

            if default_tag:
                image = repo.get_image(image_tag=default_tag)
                return image.uri
            else:
                # Fetch latest tagged
                def get_image_push_time(image: ECRImage) -> datetime:
                    if image.image_pushed_at is None:
                        raise RuntimeError(f"Couldn't get 'image_pushed_at' for: {image}")
                    return image.image_pushed_at

                images = sorted(repo.get_images("TAGGED"), key=get_image_push_time)
                return images[-1].uri
        else:
            raise ValueError(f"Could not resolve full URI for image {uri} (raw={name})")
    except Exception as e:
        msg = f"Couldn't resolve ECR image URI from {name} with error: {e}"
        logger.exception(msg)
        raise ResourceNotFoundError(msg) from e
